#!/usr/bin/perl -w

#----------------------------------------------------------------------
# runtime.pm
# 
# Runtime Perl library for web applications generated by the ELEVEN compiler.
# 
# Copyright (C) 2004 Joseph Morrison <jdm@statesafe.org>.
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details. Visit http://www.gnu.org/copyleft/gpl.html for the full text of the
# license, or you can request a copy from the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
# 
# This license allows you to use the Eleven compiler no charge under the
# condition that if you create a derived work (either based on the Eleven
# compiler or based on code generated by the Eleven compiler) and redistribute
# it, the complete source code for your application must be available and freely
# redistributable.  A commercial license for Eleven is available without this
# condition. For more information please email info@statesafe.org.
#----------------------------------------------------------------------

package runtime11;

use strict;
use warnings;

use DBI;
use POSIX qw(strftime floor);
use mod_perl;
use constant MP2 => $mod_perl::VERSION >= 1.99;

BEGIN
{
    if (MP2)
    {
	require APR::SockAddr;
	require Apache::Connection;
	require Apache::Const;
	require Apache::RequestIO;
	require Apache::RequestRec;
	Apache::Const->import (-compile => 'OK');
    }
    else
    {
	require Socket;
	require Apache::Constants;
	Apache::Constants->import ('OK');
    }
}

BEGIN
{
    use Exporter ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION     = 0.01;
    @ISA         = qw(Exporter);

    @EXPORT      = qw(&runtime11_handler
		      &gen_rowid
		      &db_read_continuation
		      &db_read_user
		      &db_read_schema
		      &var_delete
		      &var_scalar
		      &var_scalar_delete
		      &var_array_empty
		      &var_array_rowcount
		      &var_array_element
		      &var_array_deleterow
		      &var_stack_maxindex
		      &var_stack_truncate
		      &subroutine_push
		      &subroutine_pop
		      &display_state_update);

    @EXPORT_OK   = ();
    %EXPORT_TAGS = ();
}

sub runtime11_handler
{
    my $r = shift;
    my $generate_display_fn = shift;
    my $execute_program_fn = shift;
    my $db_connect = shift;
    my $db_user = shift;
    my $db_pass = shift;
    my $package_name = shift;
    my $auth_method = shift;
    my $schema_id = shift;

    #----------------------------------------------------------------------
    # Retrieve query string and POST parameters.
    #----------------------------------------------------------------------

    # Unfortunately this nice, simple way of fetching query string
    # arguments and POST parameters is now deprecated:
    # 
    #     my %parameters = ($r->args, $r->content);

    #----------------------------------------
    # Read POST parameters:
    #----------------------------------------

    my %parameters = ();
    $r->setup_client_block;
    if ($r->should_client_block)
    {
	my $data = '';
	my $buf;
	while (my $read_len = $r->get_client_block ($buf, 8192))
	{
	    if ($read_len == -1)
	    {
		db_log_message ("error while reading with get_client_block");
		last;
	    }
	    $data .= $buf;
	}
	%parameters = parse_args ($data);
    }

    #----------------------------------------
    # Read "cid" parameter from query string arguments
    # and merge with the POST parameters:
    #----------------------------------------

    my %queryargs = parse_args (scalar ($r->args));
    $parameters{'cid'} = $queryargs {'cid'};

    #----------------------------------------------------------------------
    # Connect to database.
    #----------------------------------------------------------------------

    my $dbh = DBI->connect ($db_connect, $db_user, $db_pass, { RaiseError => 0, AutoCommit => 1 });

    unless ($dbh)
    {
	$generate_display_fn->($r, "Failed to connect to database.", undef ());
	return MP2 ? Apache::OK : Apache::Constants::OK;
    }

    #----------------------------------------------------------------------
    # Process HTTP request.
    #----------------------------------------------------------------------

    # First check if the logout button was pressed. If so, start a new session. But first, check
    # if the request came from a valid continuation with a valid parent user. If so, make the
    # user's current continuation a copy of the most recent one, and delete all the others. This
    # will invalidate any screens that could be returned to via the "back" button, and force the
    # user to perform a valid login to resume the session.

    if ((defined $parameters{'action'}) && ($parameters{'action'} eq 'Logout'))
    {
	my $state_ref = db_read_continuation ($dbh, $parameters{'cid'}, 0);
	unless ($state_ref)
	{
	    start_session ($dbh, $r, "User logout. (warning: missing session information)", $package_name, $auth_method, $schema_id, $generate_display_fn, $execute_program_fn);
	}
	else
	{
	    if ($parameters{'authtoken'} ne $state_ref->{'auth_token'})
	    {
		start_session ($dbh, $r, "User logout. (warning: invalid session information)", $package_name, $auth_method, $schema_id, $generate_display_fn, $execute_program_fn);
	    }
	    else
	    {
		my $parent_uid = $state_ref->{'parent_uid'};
		if ($parent_uid != 0)
		{
		    my %next_state = %$state_ref; 
		    
		    display_state_update ($r, \%next_state, $next_state {'display_state'});

		    $next_state {'predecessor_cid'} = $state_ref->{'cid'};
		    $next_state {'back_flag'} = ($state_ref->{'next_cid'} ? 1 : 0);

		    db_insert_continuation ($dbh, \%next_state);
		    db_update_user ($dbh, $parent_uid, $next_state {'cid'});
		    db_remove_user_continuations_except ($dbh, $parent_uid, $next_state {'cid'});
		}
		start_session ($dbh, $r, "User logout.", $package_name, $auth_method, $schema_id, $generate_display_fn, $execute_program_fn);
	    }
	}
    }
    else
    {
	# If we got to this point, then the logout button was not pressed. Check if the query
	# string contains a valid continuation ID.

	my $state_ref = db_read_continuation ($dbh, $parameters{'cid'}, 0);
	unless ($state_ref)
	{
	    # No valid continuation ID found. Either the user's session expired or this is a
	    # fresh login.  In either case, start a new session.
	    
	    start_session ($dbh, $r, ($parameters{'cid'} ? "Session has expired." : "Starting new session."), $package_name, $auth_method, $schema_id, $generate_display_fn, $execute_program_fn);
	}
	else
	{
	    # Continuation found. Check the authorization token to see if it's valid. If not,
	    # start a new session. If so, check the schema id to see whether the continuation
	    # was created by this version of the application. If not, start a new
	    # session. Otherwise, resume the application based on the information in the
	    # continuation.

	    if ($parameters{'authtoken'} ne $state_ref->{'auth_token'})
	    {
		start_session ($dbh, $r, "Forced logout due to invalid attempt to connect to session.", $package_name, $auth_method, $schema_id, $generate_display_fn, $execute_program_fn);
	    }
	    else
	    {
		if ($state_ref->{'schema_id'} ne $schema_id)
		{
		    start_session ($dbh, $r, "Forced logout due to system upgrade.", $package_name, $auth_method, $schema_id, $generate_display_fn, $execute_program_fn);
		}
		else
		{
		    continue_session ($dbh, $r, \%parameters, $state_ref, $generate_display_fn, $execute_program_fn);
		}
	    }
	}
    }

    #----------------------------------------------------------------------
    # Disconnect from database.
    #----------------------------------------------------------------------

    $dbh->disconnect ();
    return MP2 ? Apache::OK : Apache::Constants::OK;
}

sub parse_args
{
    my $string = shift;
    return () unless defined $string and $string;

    return map {
        tr/+/ /;
        s/%([0-9a-fA-F]{2})/pack("C",hex($1))/ge;
        $_;
    } split /[=&;]/, $string, -1;
}

sub get_remote_addr
{
    my $r = shift;

    my $remote_port;
    my $remote_addr;

    if (MP2)
    {
	my $sockaddr = $r->connection->remote_addr;
	($remote_port, $remote_addr) = ($sockaddr->port, $sockaddr->ip_get);
    }
    else
    {
	my $sockaddr_in = $r->connection->remote_addr;
	($remote_port, $remote_addr) = Socket::sockaddr_in ($sockaddr_in);
    }

    return ($remote_addr, $remote_port);
}

my $cid_counter = 0;
sub gen_cid
{
    my $r = shift;

    # To generate a unique continuation ID, we use the IP address and port number of the client
    # and a one-second resolution timestamp. That gives us uniqueness unless the client somehow
    # generated more than one request per second from the same outgoing port number.
    # 
    # In the unlikely case that's a problem, we add in the process ID of the web server and a
    # unique counter per server. Also for good measure, we also generate three random numbers
    # between 0 and 999,999 and throw those in too.

    my ($remote_addr, $remote_port) = get_remote_addr ($r);
    my ($a, $b, $c, $d) = split (/\./, $remote_addr, 4);
    my $packed_addr = $a * 256 * 256 * 256 + $b * 256 * 256 + $c * 256 + $d;

    my $cid = sprintf ("%010u-%06u-%010u-%06u-%06u-%06u-%06u-%06u",
		       $packed_addr,
		       $remote_port,
		       time (),
		       $$,
		       $cid_counter,
		       int rand 1000000,
		       int rand 1000000,
		       int rand 1000000);

    $cid_counter ++;
    if ($cid_counter > 999999)
    {
	$cid_counter = 0;
    }

    return $cid;
}

my $rowid_counter = 0;
sub gen_rowid
{
    my $r = shift;

    # To generate a unique row ID, we do the same thing as for continuation IDs.

    my ($remote_addr, $remote_port) = get_remote_addr ($r);
    my ($a, $b, $c, $d) = split (/\./, $remote_addr, 4);
    my $packed_addr = $a * 256 * 256 * 256 + $b * 256 * 256 + $c * 256 + $d;

    my $rowid = sprintf ("%010u-%06u-%010u-%06u-%06u-%06u-%06u-%06u",
			 $packed_addr,
			 $remote_port,
			 time (),
			 $$,
			 $rowid_counter,
			 int rand 1000000,
			 int rand 1000000,
			 int rand 1000000);

    $rowid_counter ++;
    if ($rowid_counter > 999999)
    {
	$rowid_counter = 0;
    }

    return $rowid;
}

sub gen_auth_token
{
    return sprintf ("%08u%08u", int rand 100000000, int rand 100000000);
}

sub gen_timestamp
{
    my $ts = strftime ("%Y-%m-%d %H:%M:%S", localtime ());
    return $ts;
}

sub db_read_continuation
{
    my $dbh = shift;
    my $cid = shift;
    my $ignore_deleted_flag = shift;

    unless ($cid)
    {
	return undef ();
    }

    my %state = ();

    my $sth = $dbh->prepare ($ignore_deleted_flag ? "SELECT * FROM continuation WHERE cid = ?"
			                          : "SELECT * FROM continuation WHERE deleted_flag = 0 AND cid = ?");
    $sth->execute ($cid);

    my $rowcount = 0;
    while (my $ref = $sth->fetchrow_hashref ())
    {
	$state {'cid'}             = $ref->{'cid'};
 	$state {'auth_token'}      = $ref->{'auth_token'};
 	$state {'creation_time'}   = $ref->{'creation_time'};
	$state {'parent_uid'}      = $ref->{'parent_uid'};
	$state {'display_state'}   = $ref->{'display_state'};
	$state {'schema_id'}       = $ref->{'schema_id'};
	$state {'next_cid'}        = $ref->{'next_cid'};
	$state {'predecessor_cid'} = $ref->{'predecessor_cid'};
	$state {'back_flag'}       = $ref->{'back_flag'};
	$state {'deleted_flag'}    = $ref->{'deleted_flag'};
	$rowcount ++;
    }
    $sth->finish();

    if ($rowcount != 1)
    {
	return undef ();
    }

    # If we got to this point, then a valid continuation was found
    # (and was not deleted), therefore the corresponding continuation
    # data is valid and we can go ahead and read it.

    $sth = $dbh->prepare ("SELECT * FROM continuation_data WHERE parent_cid = ?");
    $sth->execute ($cid);

    while (my $ref = $sth->fetchrow_hashref ())
    {
	$state {'vars'}
	    [$ref->{'stack_frame'}]
	    [$ref->{'variable'}]
	    [$ref->{'variable_index'}] = $ref->{'value'};
    }

    $sth->finish();
    return \%state;
}

sub db_insert_continuation
{
    my $dbh = shift;
    my $state_ref = shift;

    my $sth = $dbh->prepare ("INSERT INTO continuation (cid, auth_token, creation_time, parent_uid, display_state, schema_id, next_cid, predecessor_cid, back_flag) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)");
    unless ($sth->execute ($state_ref->{'cid'},
			   $state_ref->{'auth_token'},
			   $state_ref->{'creation_time'},
			   $state_ref->{'parent_uid'},
			   $state_ref->{'display_state'},
			   $state_ref->{'schema_id'},
			   $state_ref->{'next_cid'},
			   $state_ref->{'predecessor_cid'},
			   $state_ref->{'back_flag'}))
    {
	return undef();
    }

    $sth = $dbh->prepare ("INSERT INTO continuation_data (parent_cid, stack_frame, variable, variable_index, value) VALUES (?, ?, ?, ?, ?)");

    foreach my $stack_frame (0 .. $#{$state_ref->{'vars'}})
    {
	foreach my $variable (0 .. $#{$state_ref->{'vars'}[$stack_frame]})
	{
	    foreach my $variable_index (0 .. $#{$state_ref->{'vars'}[$stack_frame][$variable]})
	    {
		# Must coerce last argument into a string by putting quotes around it. Otherwise
		# Perl will sometimes think it's a number, and other times think it's a string,
		# triggering a bug in the MySQL DBI/DBD libraries.

		# db_log_message ("writing stack frame $stack_frame, variable $variable, index $variable_index, value " .
		#                 "$state_ref->{'vars'}[$stack_frame][$variable][$variable_index]");

		$sth->execute ($state_ref->{'cid'},
			       $stack_frame,
			       $variable,
			       $variable_index,
			       "$state_ref->{'vars'}[$stack_frame][$variable][$variable_index]");
	    }
	}
    }
}

sub db_update_continuation
{
    my $dbh = shift;
    my $state_ref = shift;

    # Don't bother checking deleted_flag, since:
    # (a) this operation should never be invoked on a deleted continuation, and
    # (b) even if it is invoked on a deleted continuation, so what?

    my $sth = $dbh->prepare ("UPDATE continuation SET next_cid = ? WHERE cid = ?");
    $sth->execute ($state_ref->{'next_cid'}, $state_ref->{'cid'});
}

sub db_remove_successor_continuations
{
    my $dbh = shift;
    my $state_ref = shift;

    my $cid = $state_ref->{'next_cid'};
    my $next_cid;

    while ($cid)
    {
	$next_cid = undef ();

	# Don't bother checking deleted_flag, since:
	# (a) this operation should never be invoked on a deleted continuation, and
	# (b) even if it is invoked on a deleted continuation, so what?

	my $sth = $dbh->prepare ("SELECT next_cid FROM continuation WHERE cid = ?");
	$sth->execute ($cid);

	while (my $ref = $sth->fetchrow_hashref ())
	{
	    $next_cid = $ref->{'next_cid'};
	}
	$sth->finish();

	#----------------------------------------
	# Delete continuation with id $cid:
	#----------------------------------------
	# 
	# $sth = $dbh->prepare ("DELETE FROM continuation_data WHERE parent_cid = ?");
	# $sth->execute ($cid);
	# 
	# $sth = $dbh->prepare ("DELETE FROM continuation WHERE cid = ?");
	# $sth->execute ($cid);

	$sth = $dbh->prepare ("UPDATE continuation SET deleted_flag = 1 WHERE cid = ?");
	$sth->execute ($cid);

	$cid = $next_cid;
    }
}

#------------------------------------------------------------
# Remove all continuations for a particular user except for
# a single specified one.
#------------------------------------------------------------

sub db_remove_user_continuations_except
{
    my $dbh = shift;
    my $uid = shift;
    my $keep_cid = shift;

    # my @hitlist = ();
    # my $sth = $dbh->prepare ("SELECT cid FROM continuation WHERE parent_uid = ? AND cid <> ?");
    # 
    # $sth->execute ($uid, $keep_cid);
    # while (my $ref = $sth->fetchrow_hashref ())
    # {
    #     push (@hitlist, $ref->{'cid'});
    # }
    # $sth->finish();
    # 
    # $sth = $dbh->prepare ("DELETE FROM continuation WHERE parent_uid = ? AND cid <> ?");
    # $sth->execute ($uid, $keep_cid);
    # 
    # foreach my $cid (@hitlist)
    # {
    #     $sth = $dbh->prepare ("DELETE FROM continuation_data WHERE parent_cid = ?");
    #     $sth->execute ($cid);
    #     # db_log_message ("removed continuation $cid");
    # }

    my $sth = $dbh->prepare ("UPDATE continuation SET deleted_flag = 1 WHERE parent_uid = ? AND cid <> ?");
    $sth->execute ($uid, $keep_cid);
}

sub db_read_user
{
    my $dbh = shift;
    my $package_name = shift;
    my $login = shift;
    my $password = shift;

    my $sth = $dbh->prepare ("SELECT * FROM user WHERE package = ? AND auto_created_flag = 0 AND login = ? AND password = MD5(?)");
    $sth->execute ($package_name, $login, $password);

    my $name = undef ();
    my $uid = undef ();
    my $current_cid = undef ();

    while (my $ref = $sth->fetchrow_hashref ())
    {
	$name        = $ref->{'name'};
	$uid         = $ref->{'uid'};
 	$current_cid = $ref->{'current_cid'};
    }
 
    $sth->finish();
    
    return ($name, $uid, $current_cid);
}

sub db_update_user
{
    my $dbh = shift;
    my $uid = shift;
    my $cid = shift;

    if ($uid && $cid)
    {
	my $sth = $dbh->prepare ("UPDATE user SET current_cid = ? WHERE uid = ?");
	$sth->execute ($cid, $uid);
    }
}

sub db_read_schema
{
    my $dbh = shift;
    my $sid = shift;

    unless ($sid)
    {
	return undef ();
    }

    my @schema = ();

    my $sth = $dbh->prepare ("SELECT * FROM schema WHERE schema_id = ?");
    $sth->execute ($sid);

    while (my $ref = $sth->fetchrow_hashref ())
    {
 	$schema [$ref->{'variable'}] [$ref->{'subroutine_id'}] {'variable_name'}   = $ref->{'variable_name'};
 	$schema [$ref->{'variable'}] [$ref->{'subroutine_id'}] {'type_name'}       = $ref->{'type_name'};

 	$schema [$ref->{'variable'}] [$ref->{'subroutine_id'}] {'fields'} [$ref->{'type_index'}] {'field_name'}  = $ref->{'field_name'};
 	$schema [$ref->{'variable'}] [$ref->{'subroutine_id'}] {'fields'} [$ref->{'type_index'}] {'field_type'}  = $ref->{'field_type'};
 	$schema [$ref->{'variable'}] [$ref->{'subroutine_id'}] {'fields'} [$ref->{'type_index'}] {'field_label'} = $ref->{'field_label'};
    }

    $sth->finish();
    return \@schema;
}

sub db_log_message
{
    my $message = shift;
    system ("/usr/bin/logger -t runtime11 \"$message\"");
}

sub start_session
{
    my $dbh = shift;
    my $r = shift;
    my $status_message = shift;
    my $package_name = shift;
    my $auth_method = shift;
    my $schema_id = shift;
    my $generate_display_fn = shift;
    my $execute_program_fn = shift;

    my $state_ref;

    if ($auth_method eq 'auto')
    {
	# If authentication method is "auto", then every new session should automatically
	# generate a fresh uid:

	my $new_uid = int rand 100000000;
	my $sth = $dbh->prepare ("INSERT INTO user (package, creation_time, auto_created_flag, uid) VALUES (?, NOW(), 1, ?)");
	$sth->execute ($package_name, $new_uid);
	$sth->finish();

	# Now start the session. Initialize state with the new uid. Set display number to 2,
	# which is always the "program termination" display.  That display is immediately
	# followed by a jump to "main" (the beginning of the user program code), so call the
	# "execute" function to make the jump happen and run the user program code until the
	# first display. Then save the continuation.
	# 
	# In the future it might be cleaner to allocate a special display number that means
	# "jump to main".

	my %tmp_parameters = ();

	my $tmp_state_ref = { cid             => gen_cid ($r),
			      auth_token      => gen_auth_token (),
			      creation_time   => gen_timestamp (),
			      parent_uid      => $new_uid,
			      display_state   => 2,
			      schema_id       => $schema_id,
			      next_cid        => undef (),
			      predecessor_cid => undef (),
			      back_flag       => 0 };

	$state_ref = $execute_program_fn->($dbh, $r, \%tmp_parameters, $tmp_state_ref);
    }
    else
    {
	# For any other authentication method, jump to screen zero, which is always some sort of
	# login screeen. No need to run any user program code at this point.

	$state_ref = { cid             => gen_cid ($r),
		       auth_token      => gen_auth_token (),
		       creation_time   => gen_timestamp (),
		       parent_uid      => 0,
		       display_state   => 0,
		       schema_id       => $schema_id,
		       next_cid        => undef (),
		       predecessor_cid => undef (),
		       back_flag       => 0 };
    }

    db_insert_continuation ($dbh, $state_ref);
    $generate_display_fn->($r, $status_message, $state_ref);
}

sub continue_session
{
    my $dbh = shift;
    my $r = shift;
    my $parameters_ref = shift;
    my $state_ref = shift;
    my $generate_display_fn = shift;
    my $execute_program_fn = shift;

    my $next_state_ref = $execute_program_fn->($dbh, $r, $parameters_ref, $state_ref);

    $next_state_ref->{'predecessor_cid'} = $state_ref->{'cid'};
    $next_state_ref->{'back_flag'} = ($state_ref->{'next_cid'} ? 1 : 0);

    db_insert_continuation ($dbh, $next_state_ref);

    if ($next_state_ref->{'parent_uid'} != 0)
    {
	db_update_user ($dbh, $next_state_ref->{'parent_uid'}, $next_state_ref->{'cid'});

	if ($next_state_ref->{'resumed_session'} eq 'yes')
	{
	    db_remove_user_continuations_except ($dbh, $next_state_ref->{'parent_uid'}, $next_state_ref->{'cid'});
	    $generate_display_fn->($r, "Resuming session for user $next_state_ref->{'parent_uid'}.", $next_state_ref);
	}
	elsif ($next_state_ref->{'resumed_session'} eq 'schema_reset')
	{
	    db_remove_user_continuations_except ($dbh, $next_state_ref->{'parent_uid'}, $next_state_ref->{'cid'});
	    $generate_display_fn->($r, "Session restarted due to system upgrade.", $next_state_ref);
	}
	else
	{
	    db_remove_successor_continuations ($dbh, $state_ref);
	    $state_ref->{'next_cid'} = $next_state_ref->{'cid'};
	    db_update_continuation ($dbh, $state_ref);
	    $generate_display_fn->($r, "Updated state for user $next_state_ref->{'parent_uid'}.", $next_state_ref);
	}
    }
    else
    {
	$generate_display_fn->($r, "No user logged in.", $next_state_ref);
    }
}

sub var_delete
{
    my $state_ref = shift;
    my $stack_frame = shift;
    my $variable = shift;

    delete $state_ref->{'vars'}[$stack_frame][$variable];
}

sub var_scalar : lvalue
{
    my $state_ref = shift;
    my $stack_frame = shift;
    my $variable = shift;

    $state_ref->{'vars'}[$stack_frame][$variable][0];
}

sub var_scalar_delete
{
    my $state_ref = shift;
    my $stack_frame = shift;
    my $variable = shift;

    delete $state_ref->{'vars'}[$stack_frame][$variable][0];
}

sub var_array_empty
{
    my $state_ref = shift;
    my $stack_frame = shift;
    my $variable = shift;

    return ($#{$state_ref->{'vars'}[$stack_frame][$variable]} < 0);
}

sub var_array_rowcount
{
    my $state_ref = shift;
    my $stack_frame = shift;
    my $variable = shift;
    my $columncount = shift;

    return (1 + floor ($#{$state_ref->{'vars'}[$stack_frame][$variable]} / $columncount));
}

sub var_array_element : lvalue
{
    my $state_ref = shift;
    my $stack_frame = shift;
    my $variable = shift;
    my $columncount = shift;
    my $row = shift;
    my $col = shift;

    $state_ref->{'vars'}[$stack_frame][$variable][$row * $columncount + $col];
}

sub var_array_deleterow
{
    my $state_ref = shift;
    my $stack_frame = shift;
    my $variable = shift;
    my $columncount = shift;
    my $row = shift;

    splice (@{$state_ref->{'vars'}[$stack_frame][$variable]}, ($row * $columncount), $columncount);
}

sub var_stack_maxindex
{
    my $state_ref = shift;

    my $maxindex = $#{$state_ref->{'vars'}};
    return (($maxindex < 0) ? 0 : $maxindex);
}

sub var_stack_truncate
{
    my $state_ref = shift;
    my $stack_frame = shift;

    splice (@{$state_ref->{'vars'}}, $stack_frame + 1);
}

sub subroutine_push
{
    my $state_ref = shift;
    my $subroutine_id = shift;
    my $return_location = shift;
    my @arguments = @_;

    my $next_stack_frame = var_stack_maxindex ($state_ref) + 1;

    var_scalar ($state_ref, $next_stack_frame, 0) = $subroutine_id;
    var_scalar ($state_ref, $next_stack_frame, 1) = $return_location;
    var_scalar ($state_ref, $next_stack_frame, 2) = undef ();

    my $nextindex = 3;
    while (defined (my $arg = shift (@arguments)))
    {
	var_scalar ($state_ref, $next_stack_frame, $nextindex) = $arg;
	$nextindex ++;
    }
}

sub subroutine_pop
{
    my $state_ref = shift;
    pop (@{$state_ref->{'vars'}});
}

sub display_state_update
{
    my $r = shift;
    my $state_ref = shift;
    my $display_state = shift;

    $state_ref->{'display_state'} = $display_state;
    $state_ref->{'creation_time'} = gen_timestamp ();
    $state_ref->{'cid'}           = gen_cid ($r);
    $state_ref->{'auth_token'}    = gen_auth_token ();
    $state_ref->{'next_cid'}      = undef ();
}

1;
