;;;----------------------------------------------------------------------
;;; Eleven compiler front end (from parse tree file to intermediate code file).
;;; 
;;; Copyright (C) 2004 Joseph Morrison <jdm@statesafe.org>.
;;; 
;;; This program is free software; you can redistribute it and/or modify it under
;;; the terms of the GNU General Public License as published by the Free Software
;;; Foundation; either version 2 of the License, or (at your option) any later
;;; version.
;;; 
;;; This program is distributed in the hope that it will be useful, but WITHOUT
;;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
;;; FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
;;; details. Visit http://www.gnu.org/copyleft/gpl.html for the full text of the
;;; license, or you can request a copy from the Free Software Foundation, Inc.,
;;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;; 
;;; This license allows you to use the Eleven compiler no charge under the
;;; condition that if you create a derived work (either based on the Eleven
;;; compiler or based on code generated by the Eleven compiler) and redistribute
;;; it, the complete source code for your application must be available and freely
;;; redistributable.  A commercial license for Eleven is available without this
;;; condition. For more information please email info@statesafe.org.
;;;----------------------------------------------------------------------

(define (main)

  ;; seed random number generator
  (set! *random-state* (seed->random-state (current-time)))

  ;; command line arguments:
  ;; - filename of this compiler front end program
  ;; - compiler data directory
  ;; - input filename (parse tree)
  ;; - output filename (intermediate code filename)

  (exit (call-with-current-continuation
	 (lambda (error-exit)
	   (if (= (length (command-line)) 4)
	       (let ((input-file (list-ref (command-line) 2))
		     (output-file (list-ref (command-line) 3)))
		 (compile-file error-exit input-file output-file))
	       (compile-error error-exit #f "internal error starting compiler second stage"))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-error:
;;; 
;;; Call this procedure if an error occurs during compilation.
;;;----------------------------------------------------------------------

(define (compile-error error-exit linenumber message)
  (display "c11: ")
  (display message)
  (if linenumber
      (begin
	(display " (line ")
	(display linenumber)
	(display ")")))
  (newline)
  (error-exit 1))

;;;----------------------------------------------------------------------
;;; Procedure compile-file:
;;;----------------------------------------------------------------------

(define (compile-file error-exit input-file output-file)

  (let ((source-code-parsed (with-input-from-file input-file read)))

    (let ((options    (get-value 'options    source-code-parsed))
	  (statements (get-value 'statements source-code-parsed)))

      (let ((app-name       (compile-option options 'app_name))
	    (package-name   (compile-option options 'package_name))
	    (auth-method    (compile-option options 'auth_method))
	    (db-connect     (compile-option options 'db_connect))
	    (db-user        (compile-option options 'db_user))
	    (db-pass        (compile-option options 'db_pass))
	    (style-sheet    (compile-option options 'style_sheet))
	    (fav-icon       (compile-option options 'fav_icon))
	    (display-header (compile-option options 'display_header)))

	(let ((main-target (gen-temp "main"))
	      (schema-id   (gen-schema-id)))

	  (call-with-values (lambda () (compile-stmtlist error-exit
							 (new-state)
							 (new-symtab)
							 (new-symtab)
							 statements
							 `(temp ,main-target)))
	    (lambda (new-state
		     new-active-symtab
		     new-full-symtab
		     new-instructions
		     new-post-state-instructions
		     new-post-control-instructions
		     new-subroutine-instructions)

	      (let ((instructions `(
				    start
				    (display 0 "login" normal)
				    (authenticate)
				    (display 1 "bad login" normal)
				    (goto start)
				    main
				    (check-schema)
				    (declare-temp ,main-target)
				    ,@new-instructions
				    end
				    (comment "Program is about to terminate. Truncate stack in case we called exit from within a subroutine.")
				    (stack-truncate (const 0))
				    (display 2 "end session" normal)
				    (goto main)
				    ,@new-subroutine-instructions
				    )))
		
		(with-output-to-file output-file
		  (lambda ()
		    (display ";;;----------------------------------------------------------------------")
		    (newline)
		    (display ";;; Intermediate code file generated by the Eleven compiler stage 2 (front end)")
		    (newline)
		    (display ";;;----------------------------------------------------------------------")
		    (newline)
		    (newline)
		    (write `((app-name       ,app-name)
			     (package-name   ,package-name)
			     (auth-method    ,auth-method)
			     (db-connect     ,db-connect)
			     (db-user        ,db-user)
			     (db-pass        ,db-pass)
			     (style-sheet    ,style-sheet)
			     (fav-icon       ,fav-icon)
			     (display-header ,display-header)
			     (schema-id      ,schema-id)
			     (schema-entries (,@(symtab-schema-entries new-full-symtab)
					      ,@(display-schema-entries instructions)))
			     (transient-vars ,(symtab-transient-vars new-full-symtab))
			     (instructions   ,instructions)))
		    (newline)))))))))))

;;;----------------------------------------------------------------------
;;; Read the requested value from the intermediate code.
;;;----------------------------------------------------------------------

(define (get-value key alist)
  (let ((value (assq key alist)))
    (if value
	(cadr value)
	(begin
	  (display "c11: missing \"")
	  (display key)
	  (display "\" value in compiler third stage")
	  (newline)
	  (exit 1)))))
	
;;;----------------------------------------------------------------------
;;; Procedure compile-option:
;;; 
;;; Look up an Eleven compiler option. Supply a default if the input Eleven
;;; program didn't specify a value.
;;;----------------------------------------------------------------------

(define (compile-option options option-name)
  (let ((value (assq option-name `(,@options
				   (app_name       "Eleven application")
				   (package_name   "eleven_application")
				   (auth_method    "auto")
				   (db_connect     "DBI:mysql:database=eleven;host=localhost")
				   (db_user        "www")
				   (db_pass        "wwwpass")
				   (style_sheet    "http://eleven.sourceforge.net/style/eleven.css")
				   (fav_icon       "http://eleven.sourceforge.net/favicon.ico")
				   (display_header "yes")))))
    (if value (cadr value) "<undefined>")))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmtlist:
;;; 
;;; Compile a list of statements.
;;; 
;;; (compile-stmtlist error-exit state active-symtab full-symtab stmtlist target)
;;; 
;;; Arguments:
;;; 
;;; error-exit    : Exit procedure (for immediate termination in case of error)
;;; state         : Compiler state (inside display command, inside subroutine, etc.)
;;; active-symtab : Active list of variable and type definitions for compilation
;;; full-symtab   : Complete list of variable and type definitions for documentation
;;; stmtlist      : List of statements to be compiled
;;; target        : Where the compiled statements should store their result value
;;; 
;;; Return values:
;;; 
;;; new-state                     : Updated state after compiling statements
;;; new-active-symtab             : New active symtab after compiling statements
;;; new-full-symtab               : New full symtab after compiling statements
;;; new-instructions              : Primary generated instructions from compiling statement list
;;; new-post-state-instructions   : Additional generated instructions to save state (to execute after "display" statement)
;;; new-post-control-instructions : Additional generated instructions to check for control transfers (to execute after "display" statement)
;;; new-subroutine-instructions   : Additional generated instructions (subroutines to collect at end of program)
;;;----------------------------------------------------------------------

(define (compile-stmtlist error-exit state active-symtab full-symtab stmtlist target)

  (let ((instructions              '())
	(post-state-instructions   '())
	(post-control-instructions '())
	(subroutine-instructions   '()))

    (for-each (lambda (stmt)
		(call-with-values (lambda () (compile-stmt error-exit state active-symtab full-symtab stmt target))
		  (lambda (new-state
			   new-active-symtab
			   new-full-symtab
			   new-instructions
			   new-post-state-instructions
			   new-post-control-instructions
			   new-subroutine-instructions)

		    (set! state         new-state)
		    (set! active-symtab new-active-symtab)
		    (set! full-symtab   new-full-symtab)

		    (set! instructions 		    `(,@instructions              ,@new-instructions))
		    (set! post-state-instructions   `(,@post-state-instructions   ,@new-post-state-instructions))
		    (set! post-control-instructions `(,@post-control-instructions ,@new-post-control-instructions))
		    (set! subroutine-instructions   `(,@subroutine-instructions   ,@new-subroutine-instructions)))))

	      stmtlist)

    (values state
	    active-symtab
	    full-symtab
	    instructions
	    post-state-instructions
	    post-control-instructions
	    subroutine-instructions)))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt:
;;;----------------------------------------------------------------------

(define (compile-stmt error-exit state active-symtab full-symtab stmt target)
  (if (pair? stmt)
      (cond ((eq? (car stmt) 'define-type)            (compile-stmt-define-type  	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'define-variables)       (compile-stmt-define-variables  	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'define-variable-assign) (compile-stmt-define-variable-assign error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'define-subroutine)      (compile-stmt-define-subroutine 	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'return)                 (compile-stmt-return            	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'exit)                   (compile-stmt-exit            	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'break)                  (compile-stmt-break            	   error-exit state active-symtab full-symtab stmt target))
	    ((eq? (car stmt) 'display)                (compile-stmt-display           	   error-exit state active-symtab full-symtab stmt target))
	    ((eq? (car stmt) 'call)                   (compile-stmt-call           	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'assign)                 (compile-stmt-assign            	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'if)                     (compile-stmt-if                	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'ifelse)                 (compile-stmt-ifelse            	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'while)                  (compile-stmt-while             	   error-exit state active-symtab full-symtab stmt target))
            ((eq? (car stmt) 'preparemission)         (compile-stmt-preparemission     	   error-exit state active-symtab full-symtab stmt target))
	    (else (compile-error error-exit #f (string-append "internal error - " (symbol->string (car stmt)) " function not implemented yet"))))
      (compile-error error-exit #f (string-append "internal error - function not implemented yet"))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-define-type:
;;;----------------------------------------------------------------------

(define (compile-stmt-define-type error-exit state active-symtab full-symtab stmt target)
  (let ((linenumber (cadr stmt))
	(name (caddr stmt))
	(type (cadddr stmt)))
    (let ((sym (new-sym-typedef name type)))
      (values state
	      (extend-symtab active-symtab sym)
	      (extend-symtab full-symtab sym)
	      '()
	      '()
	      '()
	      '()))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-define-variables:
;;;----------------------------------------------------------------------

(define (compile-stmt-define-variables error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber  (cadr stmt))
	(varnamelist (caddr stmt))
	(mode (cadddr stmt))
	(type (cadr (cdddr stmt)))
	(instructions '()))

    (for-each (lambda (varname)
		(let ((actualtype (search-symtab-type active-symtab type error-exit linenumber)))

		  (cond ((eq? mode 'statesafe) (let ((sym (if (state-in-subroutine? state)
							      (new-sym-auto-statesafe (state-subroutine-id state) (state-subroutine-name state) varname `(statesafe ,actualtype))
							      (new-sym-static-statesafe varname `(statesafe ,actualtype)))))

						 (set! active-symtab (extend-symtab active-symtab sym))
						 (set! full-symtab   (extend-symtab full-symtab sym))
						 (set! instructions `(,@instructions
								      (comment ,(string-append "Initialize variable \"" (symbol->string varname) "\":"))
								      ,@(if (or (sym-static-statesafe-var? sym)
										(sym-static-statesafe-table? sym)
										(sym-static-statesafe-enum? sym)
										(sym-static-statesafe-launcher? sym))
									    `((delete-var (const 0) (const ,(sym-statesafe-index sym))))
									    (let ((temp (gen-temp "stackframe")))
									      `((declare-temp ,temp)
										(stack-maxindex (temp ,temp))
										(delete-var (temp ,temp) (const ,(sym-statesafe-index sym))))))))))

			((eq? mode 'transient) (if (and (symbol? actualtype) (eq? actualtype 'var))
						   (let ((sym (if (state-in-subroutine? state)
								  (new-sym-auto-transient (state-subroutine-id state) (state-subroutine-name state) varname `(transient ,actualtype))
								  (new-sym-static-transient varname `(transient ,actualtype)))))
				  
						     (set! active-symtab (extend-symtab active-symtab sym))
						     (set! full-symtab   (extend-symtab full-symtab sym))
						     (set! instructions `(,@instructions
									  (declare-temp ,(sym-transient-varname sym)))))
						   (compile-error error-exit linenumber "transient storage class not supported for this type")))
						   
			((eq? mode 'user)
			 (compile-error error-exit linenumber "user storage class not implemented yet"))

			((eq? mode 'global)
			 (compile-error error-exit linenumber "global storage class not implemented yet"))

			(else (compile-error error-exit linenumber "bad type")))))
	      
	      varnamelist)

    (values state active-symtab full-symtab instructions '() '() '())))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-define-variable-assign:
;;;----------------------------------------------------------------------

(define (compile-stmt-define-variable-assign error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(varname (caddr stmt))
	(expr (cadddr stmt))
	(type (cddddr stmt)))

    ;; first compile variable declaration

    (call-with-values (lambda () (compile-stmt error-exit state active-symtab full-symtab `(define-variables ,linenumber (,varname) ,@type) target))
      (lambda (decl-state
	       decl-active-symtab
	       decl-full-symtab
	       decl-instructions
	       decl-post-state-instructions
	       decl-post-control-instructions
	       decl-subroutine-instructions)

	;; now compile assignment

	(call-with-values (lambda () (compile-stmt error-exit state decl-active-symtab decl-full-symtab `(assign ,linenumber ,varname ,expr) target))
	  (lambda (assign-state
		   assign-active-symtab
		   assign-full-symtab
		   assign-instructions
		   assign-post-state-instructions
		   assign-post-control-instructions
		   assign-subroutine-instructions)

	    ;; return the updated symbol tables and the combined set of instructions

	    (values state
		    assign-active-symtab
		    assign-full-symtab 
		    `(,@decl-instructions
		      ,@assign-instructions)
		    '()
		    '()
		    '())))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-define-subroutine:
;;;----------------------------------------------------------------------

(define (compile-stmt-define-subroutine error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(subroutine-name (caddr stmt))
	(arglist  (cadddr stmt))
	(stmtlist (car (cddddr stmt))))

    (if (or (state-in-display? state)
	    (state-in-subroutine? state))
	(compile-error error-exit linenumber "cannot define subroutine within subroutine or display"))
  
    (let ((sym (new-sym-subroutine subroutine-name)))

      (set! active-symtab (extend-symtab active-symtab sym))
      (set! full-symtab (extend-symtab full-symtab sym))

      (let ((temp-active-symtab active-symtab))

	;; Push arguments into environment.

	(for-each (lambda (arg)

		    (let ((arg-type (search-symtab-type active-symtab (car arg) error-exit linenumber))
			  (arg-name (cadr arg)))

		      (cond ((eq? arg-type 'var)
			     (let ((sym (new-sym-auto-statesafe (sym-subroutine-id sym)
								subroutine-name
								arg-name
								(list 'statesafe arg-type))))
			       (set! temp-active-symtab (extend-symtab temp-active-symtab sym))
			       (set! full-symtab   (extend-symtab full-symtab sym))))

			    (else (compile-error error-exit linenumber "subroutine arguments must all be var type")))))

		  arglist)

	(call-with-values (lambda () (compile-stmtlist error-exit
						       (extend-state state (state-in-subroutine (sym-subroutine-id sym) subroutine-name))
						       temp-active-symtab
						       full-symtab
						       stmtlist
						       `(temp ,target)))
	  (lambda (new-state
		   new-active-symtab
		   new-full-symtab
		   new-instructions
		   new-post-state-instructions
		   new-post-control-instructions
		   new-subroutine-instructions)

	    ;; After compiling the subroutine, we want the active symtab to be extended with the
	    ;; subroutine name, but not any of the junk left over from compiling the
	    ;; subroutine. But the full symtab should have everything.

	    (values state active-symtab new-full-symtab '() '() '() `((comment ,(string-append "Start of subroutine \"" (symbol->string subroutine-name) "\":"))
								      ,(sym-subroutine-label sym)
								      ,@new-instructions
								      (return)
								      ,@new-subroutine-instructions
								      (comment ,(string-append "End of subroutine \"" (symbol->string subroutine-name) "\":"))))))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-return:
;;;----------------------------------------------------------------------

(define (compile-stmt-return error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(value (caddr stmt)))

    (if (or (not (state-in-subroutine? state))
	    (state-in-display? state))
	(compile-error error-exit linenumber "cannot use return statement outside subroutine or inside display")

	(let ((topstack  (gen-temp "topstack"))
	      (returnval (gen-temp "returnval")))

	  (let ((instructions (compile-expr-general error-exit linenumber state active-symtab full-symtab value `(temp ,returnval))))

	    (values state
		    active-symtab
		    full-symtab
		    `((declare-temp ,returnval)
		      ,@instructions
		      (comment "Write return value to slot 2 of the top activation record:")
		      (declare-temp ,topstack)
		      (stack-maxindex (temp ,topstack))
		      (write-scalar (temp ,returnval) (temp ,topstack) (const 2))
		      (return))
		    '()
		    '()
		    '()))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-exit:
;;;----------------------------------------------------------------------

(define (compile-stmt-exit error-exit state active-symtab full-symtab stmt target)
  (let ((linenumber (cadr stmt)))

    (if (state-in-display? state)
	(compile-error error-exit linenumber "cannot use exit statement within a display"))
    
    (values state
	    active-symtab
	    full-symtab
	    `((comment "Exit statement - jump to end of program:")
	      (goto end))
	    '()
	    '()
	    '())))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-break:
;;;----------------------------------------------------------------------

(define (compile-stmt-break error-exit state active-symtab full-symtab stmt target)
  (let ((linenumber (cadr stmt)))

    (if (not (state-in-loop? state))
	(compile-error error-exit linenumber "cannot use break statement outside loop"))
    
    (values state
	    active-symtab
	    full-symtab
	    `((comment "Break statement - jump to end of loop:")
	      (goto ,(state-loop-endlabel state)))
	    '()
	    '()
	    '())))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-display:
;;;----------------------------------------------------------------------

(define (compile-stmt-display error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber     (cadr stmt))
	(display-name   (caddr stmt))
	(display-option (cadddr stmt))
	(stmtlist       (cddddr stmt))
	(displaytemp    (gen-temp "display"))
	(display-id     (gen-display-id)))

    (if (state-in-display? state)
	(compile-error error-exit linenumber "cannot define display within another display")
	
	;; compile display statements with fresh state (forget about being in a subroutine, loop, etc.)
	(call-with-values (lambda () (compile-stmtlist error-exit (extend-state (new-state) (state-in-display display-id)) active-symtab full-symtab stmtlist `(temp ,displaytemp)))
	  (lambda (new-state
		   new-active-symtab
		   new-full-symtab
		   new-instructions
		   new-post-state-instructions
		   new-post-control-instructions
		   new-subroutine-instructions)

	    (let ((instructions `((display ,display-id
					   ,display-name
					   ,display-option
					   (declare-temp ,displaytemp)
					   ,@new-instructions)
				  ,@new-post-state-instructions
				  ,@new-post-control-instructions)))

	      ;; After compiling the display body, we want to revert to the original state and active
	      ;; symtab.  (For example, we don't want variables defined in the display to leak to
	      ;; the outside.)

	      (values state active-symtab new-full-symtab instructions '() '() new-subroutine-instructions)))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-call:
;;; 
;;; Check for special cases "print" and "edit".
;;; Otherwise compile generic subroutine call.
;;;----------------------------------------------------------------------

(define (compile-stmt-call error-exit state active-symtab full-symtab stmt target)
  (let ((linenumber (cadr stmt))
	(displaystyle (caddr stmt))
	(subroutinename (cadddr stmt)))

    (cond ((eq? subroutinename 'print)
	   (compile-stmt-print error-exit state active-symtab full-symtab stmt target))

	  ((eq? subroutinename 'image)
	   (compile-stmt-image error-exit state active-symtab full-symtab stmt target))

	  ((eq? subroutinename 'edit)
	   (compile-stmt-edit error-exit state active-symtab full-symtab stmt target))

	  ((eq? subroutinename 'include)
	   (compile-stmt-include error-exit state active-symtab full-symtab stmt target))

	  ;; Subroutine statements (i.e. not used in an expression context) other than print,
	  ;; image, edit, and use should be treated just like subroutine expressions with the
	  ;; return value discarded.

	  (else
	   (let ((temp (gen-temp "discard")))

	     (if (not (null? displaystyle))
		 (compile-error error-exit linenumber "cannot specify display style for non display-related subroutine call")

		 (let ((instructions (compile-expr-general-application-subroutine-call error-exit
										       linenumber
										       state
										       active-symtab
										       full-symtab
										       (cdddr stmt)
										       `(temp ,temp))))
		   (values state
			   active-symtab
			   full-symtab
			   `((declare-temp ,temp)
			     ,@instructions)
			   '()
			   '()
			   '()))))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-print:
;;;----------------------------------------------------------------------

(define (compile-stmt-print error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber   (cadr stmt))
	(displaystyle (caddr stmt))
	(itemlist     (cddddr stmt)))

    (let ((style-id (cond ((null? displaystyle) #f)
			  ((null? (cadr displaystyle)) #f)
			  (else (cadr displaystyle)))))

      (let ((instructions `((comment "Start of print statement:")
			    ,(if style-id
				 `(html ,displaystyle (const ,(string-append "                    <tr class=\"" style-id "\">")))
				 `(html ,displaystyle (const "                    <tr>")))
			    (html-newline ,displaystyle)
			    (html ,displaystyle (const "                      <td class=\"mainDisplayCell\" colspan=\"2\">"))
			    (html-newline ,displaystyle)))

	    (post-control-instructions '())
	    (in-paragraph #f))

	(if (not (state-in-display? state))
	    (compile-error error-exit linenumber "cannot use print statement outside display"))

	(for-each (lambda (item)
		    (let ((new-instructions

			   (cond ((expr-self-evaluating? item)

				  ;; For a self-evaluating item (number or string) generate the printing instructions.  Then if we
				  ;; are already inside a paragraph, emit the instructions directly. Otherwise prepend a <p> tag and
				  ;; set the in-paragraph flag.

				  (let ((print-instructions (compile-expr-print-self-evaluating error-exit linenumber displaystyle state active-symtab full-symtab item target)))
				    (if in-paragraph
					print-instructions
					(begin
					  (set! in-paragraph #t)
					  `((html ,displaystyle (const "                        <p>"))
					    ,@print-instructions)))))

				 ((expr-variable? item)
				  (let ((sym (search-symtab active-symtab item)))

				    (cond ((not sym)
					   (compile-error error-exit linenumber (string-append "unknown variable \"" (symbol->string item) "\"")))
					  
					  ;; For a static statesafe variable, generate the printing instructions.  Then if we are
					  ;; already inside a paragraph, emit the instructions directly. Otherwise prepend a <p> tag
					  ;; and set the in-paragraph flag.

					  ((sym-static-statesafe-var? sym)
					   (let ((print-instructions (compile-expr-print-static-statesafe-var error-exit linenumber displaystyle state active-symtab full-symtab sym target)))
					     (if in-paragraph
						 print-instructions
						 (begin
						   (set! in-paragraph #t)
						   `((html ,displaystyle (const "                        <p>"))
						     ,@print-instructions)))))
					  
					  ;; For an automatic statesafe variable, generate the printing instructions.  Then if we
					  ;; are already inside a paragraph, emit the instructions directly. Otherwise prepend a <p>
					  ;; tag and set the in-paragraph flag.

					  ((sym-auto-statesafe-var? sym)
					   (let ((print-instructions (compile-expr-print-auto-statesafe-var error-exit linenumber displaystyle state active-symtab full-symtab sym target)))
					     (if in-paragraph
						 print-instructions
						 (begin
						   (set! in-paragraph #t)
						   `((html ,displaystyle (const "                        <p>"))
						     ,@print-instructions)))))

					  ;; For a statesafe table, generate the printing instructions.  Then if we are inside a
					  ;; paragraph, unset the in-paragraph flag, close the paragraph, and print the
					  ;; table. Otherwise, just print the table directly.

					  ((or (sym-static-statesafe-table? sym) (sym-auto-statesafe-table? sym))
					   (let ((print-instructions (compile-expr-print-statesafe-table error-exit linenumber displaystyle state active-symtab full-symtab sym target)))
					     (if in-paragraph
						 (begin
						   (set! in-paragraph #f)
						   `((html ,displaystyle (const "</p>"))
						     (html-newline ,displaystyle)
						     ,@print-instructions))
						 print-instructions)))

					  ;; For an enumeration variable, generate the printing instructions.  Then if we are
					  ;; already inside a paragraph, emit the instructions directly. Otherwise prepend a <p> tag
					  ;; and set the in-paragraph flag.

					  ((or (sym-static-statesafe-enum? sym) (sym-auto-statesafe-enum? sym))
					   (let ((print-instructions (compile-expr-print-statesafe-enum error-exit linenumber displaystyle state active-symtab full-symtab sym target)))
					     (if in-paragraph
						 print-instructions
						 (begin
						   (set! in-paragraph #t)
						   `((html ,displaystyle (const "                        <p>"))
						     ,@print-instructions)))))

					  ;; For a launcher variable, generate the printing instructions.  Then if we are inside a
					  ;; paragraph, unset the in-paragraph flag, close the paragraph, and print the
					  ;; launcher. Otherwise, just print the launcher directly.

					  ((or (sym-static-statesafe-launcher? sym) (sym-auto-statesafe-launcher? sym))
					   (call-with-values (lambda ()
							       (compile-expr-print-statesafe-launcher error-exit linenumber displaystyle state active-symtab full-symtab sym target))
					     (lambda (new-instructions
						      new-post-control-instructions)
					       (set! post-control-instructions `(,@post-control-instructions
										 ,@new-post-control-instructions))
					       (if in-paragraph
						   (begin
						     (set! in-paragraph #f)
						     `((html ,displaystyle (const "</p>"))
						       (html-newline ,displaystyle)
						       ,@new-instructions))
						   new-instructions))))

					  ;; For a static statesafe variable, generate the printing instructions.  Then if we are
					  ;; already inside a paragraph, emit the instructions directly. Otherwise prepend a <p> tag
					  ;; and set the in-paragraph flag.

					  ((or (sym-static-statesafe-text? sym) (sym-auto-statesafe-text? sym))
					   (let ((print-instructions (compile-expr-print-statesafe-text error-exit linenumber displaystyle state active-symtab full-symtab sym target)))
					     (if in-paragraph
						 (begin
						   (set! in-paragraph #f)
						   `((html ,displaystyle (const "</p>"))
						     (html-newline ,displaystyle)
						     ,@print-instructions))
						 print-instructions)))

					  ;; For a transient variable, generate the printing instructions.  Then if we are already
					  ;; inside a paragraph, emit the instructions directly. Otherwise prepend a <p> tag and set
					  ;; the in-paragraph flag.

					  ((or (sym-static-transient-var? sym) (sym-auto-transient-var? sym))
					   (let ((print-instructions (compile-expr-print-transient-var error-exit linenumber displaystyle state active-symtab full-symtab sym target)))
					     (if in-paragraph
						 print-instructions
						 (begin
						   (set! in-paragraph #t)
						   `((html ,displaystyle (const "                        <p>"))
						     ,@print-instructions)))))

					  (else
					   (compile-error error-exit linenumber (string-append "variable \"" (symbol->string item) "\" has invalid or unimplemented type for printing"))))))

				 ;; For a general expression, generate the printing instructions.  Then if we are already inside a
				 ;; paragraph, emit the instructions directly. Otherwise prepend a <p> tag and set the in-paragraph
				 ;; flag.

				 ((expr-application? item)
				  (let ((print-instructions (compile-expr-print-application error-exit linenumber displaystyle state active-symtab full-symtab item target)))
				    (if in-paragraph
					print-instructions
					(begin
					  (set! in-paragraph #t)
					  `((html ,displaystyle (const "                        <p>"))
					    ,@print-instructions)))))

				 (else
				  (compile-error error-exit linenumber "invalid print argument")))))

		      (set! instructions `(,@instructions ,@new-instructions))))

		  itemlist)

	;; If we are inside a paragraph, close it now.

	(if in-paragraph
	    (set! instructions `(,@instructions
				 (html ,displaystyle (const "</p>"))
				 (html-newline ,displaystyle))))

	(set! instructions `(,@instructions
			     (html ,displaystyle (const "                      </td>"))
			     (html-newline ,displaystyle)
			     (html ,displaystyle (const "                    </tr>"))
			     (html-newline ,displaystyle)
			     (comment "End of print statement.")))
				  
	(values state active-symtab full-symtab instructions '() post-control-instructions '())))))

(define (compile-expr-print-self-evaluating error-exit linenumber displaystyle state active-symtab full-symtab item target)
  `((html ,displaystyle (const ,item))))

(define (compile-expr-print-static-statesafe-var error-exit linenumber displaystyle state active-symtab full-symtab sym target)
  (let ((temp (gen-temp (symbol->string (sym-name sym)))))
    `((comment "Print statesafe var (global; use stack frame 0):")
      (declare-temp ,temp)
      (read-scalar (temp ,temp) (const 0) (const ,(sym-statesafe-index sym)))
      (html ,displaystyle (temp ,temp)))))

(define (compile-expr-print-auto-statesafe-var error-exit linenumber displaystyle state active-symtab full-symtab sym target)
  (let ((temp (gen-temp (symbol->string (sym-name sym))))
	(stackframe-temp (gen-temp "stackframe")))

    `((comment "Print statesafe var (in subroutine; use top of stack):")
      (declare-temp ,stackframe-temp)
      (stack-maxindex (temp ,stackframe-temp))
      (declare-temp ,temp)
      (read-scalar (temp ,temp) (temp ,stackframe-temp) (const ,(sym-statesafe-index sym)))
      (html ,displaystyle (temp ,temp)))))

(define (compile-expr-print-statesafe-table error-exit linenumber displaystyle state active-symtab full-symtab sym target)

  (let ((table-varindex (sym-statesafe-index sym))
	(table-fields (sym-statesafe-type-fields sym))
	(table-stackframe-temp (gen-temp "stackframe")))
	
    `((declare-temp ,table-stackframe-temp)

      ,(if (sym-static-statesafe-table? sym)
	   `(comment ,(string-append "Print statesafe table \"" (symbol->string (sym-name sym)) "\" (global; use stack frame 0):"))
	   `(comment ,(string-append "Print statesafe table \"" (symbol->string (sym-name sym)) "\" (in subroutine; use top of stack):")))

      ;; Store the table's stack frame in temporary variable for use
      ;; throughout this "print table" subroutine.

      ,(if (sym-static-statesafe-table? sym)
	   `(assign (temp ,table-stackframe-temp) (const 0))
	   `(stack-maxindex (temp ,table-stackframe-temp)))

      ;; If the table is empty, print a message to that
      ;; effect. Otherwise, display the entire table.

      ,@(let ((isempty-temp    (gen-temp  "isempty"))
	      (noentries-label (gen-label "noentries"))
	      (finished-label  (gen-label "finished")))
	  
	  `((declare-temp ,isempty-temp)

	    (array-empty (temp ,isempty-temp) (temp ,table-stackframe-temp) (const ,table-varindex))
	    (goto-if-true ,noentries-label (temp ,isempty-temp))

	    ;; If we've gotten to this point, the table is not empty.
	    ;; Print it out in full.

	    ,@(let ((currentrow-temp  (gen-temp  "currentrow"))
		    (rowcount-temp    (gen-temp  "rowcount"))
		    (finished-temp    (gen-temp  "finished"))
		    (looptop-label    (gen-label "looptop"))
		    (loopbottom-label (gen-label "loopbottom")))

		`((declare-temp ,currentrow-temp)
		  (declare-temp ,rowcount-temp)
		  (declare-temp ,finished-temp)

		  (html ,displaystyle (const ,(string-append "                        <table summary=\"" (symbol->string (sym-name sym)) "\">")))
		  (html-newline ,displaystyle)

		  ;; Print out table header (column labels).

		  (html ,displaystyle (const "                          <tr>"))
		  (html-newline ,displaystyle)

		  ,@(let ((instructions '()))
		      (for-each (lambda (field)
				  (set! instructions `(,@instructions
						       (html ,displaystyle (const ,(string-append "                            <td class=\"tableHeaderCell\">&nbsp;" (field-label field) "&nbsp;</td>")))
						       (html-newline ,displaystyle))))
				table-fields)
		      instructions)

		  (html ,displaystyle (const "                          </tr>"))
		  (html-newline ,displaystyle)

		  ;; Loop over all the rows, then within each one loop
		  ;; over all the columns. Read each value into a
		  ;; temporary variable and print it.

		  (assign (temp ,currentrow-temp) (const 0))

		  ,looptop-label

		  (array-rowcount (temp ,rowcount-temp) (temp ,table-stackframe-temp) (const ,table-varindex) (const ,(fieldlist-colcount table-fields)))
		  (number-equal (temp ,finished-temp) (temp ,currentrow-temp) (temp ,rowcount-temp))
		  (goto-if-true ,loopbottom-label (temp ,finished-temp))

		  (html ,displaystyle (const "                          <tr>"))
		  (html-newline ,displaystyle)

		  ,@(let ((cell-temp (gen-temp "temp")))
		      (let ((tempcode `((declare-temp ,cell-temp))))

			(do ((i 0 (+ i 1)))
			    ((= i (length table-fields)) tempcode)
			  (set! tempcode `(,@tempcode
					   (html ,displaystyle (const "                            <td class=\"tableCell\">&nbsp;"))
					   (read-array-element (temp ,cell-temp)
							       (temp ,table-stackframe-temp)
							       (const ,table-varindex)
							       (const ,(fieldlist-colcount table-fields))
							       (temp ,currentrow-temp)
							       (const ,(+ i 1)))
					   (html ,displaystyle (temp ,cell-temp))
					   (html ,displaystyle (const "&nbsp;</td>"))
					   (html-newline ,displaystyle))))))

		  (html ,displaystyle (const "                          </tr>"))
		  (html-newline ,displaystyle)

		  (increment (temp ,currentrow-temp) (const 1))
		  (goto ,looptop-label)

		  ,loopbottom-label

		  (html ,displaystyle (const "                        </table>"))
		  (html-newline ,displaystyle)))

	    (goto ,finished-label)
      
	    ,noentries-label

	    ;; If we've gotten to this point, the table is empty.
	    ;; Print a message to that effect.

	    (html ,displaystyle (const "                        <p>No entries.</p>"))
	    (html-newline ,displaystyle)

	    ,finished-label)))))

(define (compile-expr-print-statesafe-enum error-exit linenumber displaystyle state active-symtab full-symtab sym target)
  (let ((enum-varindex (sym-statesafe-index sym))
	(enum-fields   (sym-statesafe-type-fields sym))
	(enum-stackframe-temp (gen-temp "stackframe")))

    `((declare-temp ,enum-stackframe-temp)

      ,(if (sym-static-statesafe-enum? sym)
	   `(comment ,(string-append "Print statesafe enum \"" (symbol->string (sym-name sym)) "\" (global; use stack frame 0):"))
	   `(comment ,(string-append "Print statesafe enum \"" (symbol->string (sym-name sym)) "\" (in subroutine; use top of stack):")))

      ;; Store the enum's stack frame in temporary variable for use
      ;; throughout this "print enum" subroutine.

      ,(if (sym-static-statesafe-enum? sym)
	   `(assign (temp ,enum-stackframe-temp) (const 0))
	   `(stack-maxindex (temp ,enum-stackframe-temp)))

      ;; Generate code to fetch the stored enum value, then compare it
      ;; to each of the possible enum values in the type definition.
      ;; When the matching value is found, print the corresponding
      ;; enum label.

      ,@(let ((enum-value-temp (gen-temp (symbol->string (sym-name sym))))
	      (stringequal-temp (gen-temp "stringequal")))

	  `((declare-temp ,enum-value-temp)
	    (declare-temp ,stringequal-temp)

	    (read-scalar (temp ,enum-value-temp) (temp ,enum-stackframe-temp) (const ,enum-varindex))

	    ,@(let ((instructions '()))

		(for-each (lambda (field)
			    (let ((foundmatch-label (gen-label "foundmatch"))
				  (continue-label (gen-label "continue")))
			      (set! instructions `(,@instructions
						   (string-equal (temp ,stringequal-temp) (temp ,enum-value-temp) (const ,(symbol->string (field-name field))))
						   (goto-if-true ,foundmatch-label (temp ,stringequal-temp))
						   (goto ,continue-label)
						   ,foundmatch-label
						   (html ,displaystyle (const ,(field-label field)))
						   ,continue-label))))
		    enum-fields)
		instructions))))))

(define (compile-expr-print-statesafe-launcher error-exit linenumber displaystyle state active-symtab full-symtab sym target)

  (let ((launcher-varindex (sym-statesafe-index sym))
	(launcher-stackframe-temp (gen-temp "stackframe"))
	(launcher-action (gen-action-selector-value "launcher")))
	
    (values `((comment "Display launcher choices:")

	      (declare-temp ,launcher-stackframe-temp)

	      ,(if (sym-static-statesafe-launcher? sym)
		   `(assign (temp ,launcher-stackframe-temp) (const 0))
		   `(stack-maxindex (temp ,launcher-stackframe-temp)))

	      ,@(let ((isempty-temp     (gen-temp  "isempty"))
		      (noentries-label  (gen-label "noentries"))
		      (done-label       (gen-label "done"))
		      (currentrow-temp  (gen-temp  "currentrow"))
		      (rowcount-temp    (gen-temp  "rowcount"))
		      (finished-temp    (gen-temp  "finished"))
		      (looptop-label    (gen-label "looptop"))
		      (loopbottom-label (gen-label "loopbottom")))

		  `((declare-temp ,isempty-temp)
		    (declare-temp ,currentrow-temp)
		    (declare-temp ,rowcount-temp)
		    (declare-temp ,finished-temp)

		    (array-empty (temp ,isempty-temp) (temp ,launcher-stackframe-temp) (const ,launcher-varindex))
		    (goto-if-true ,noentries-label (temp ,isempty-temp))

		    ,@(if (or (null? displaystyle)
			      (eq? (car displaystyle) 'header)
			      (eq? (car displaystyle) 'main)
			      (eq? (car displaystyle) 'footer))
			  `((html ,displaystyle (const "                        <p>"))
			    (html-newline ,displaystyle)
			    (html ,displaystyle (const "                        [ ")))
			  `())

		    ;; In a launcher object, slots 0, 1, and 2 are reserved. Launcher choices start in slot 3.

		    (comment "Launcher slots 0, 1, and 2 are reserved for the mission stack frame, user selection, and mission id. Labels start in slot 3.")
		    (assign (temp ,currentrow-temp) (const 3))

		    ,looptop-label

		    (array-rowcount (temp ,rowcount-temp) (temp ,launcher-stackframe-temp) (const ,launcher-varindex) (const 1))
		    (number-equal (temp ,finished-temp) (temp ,currentrow-temp) (temp ,rowcount-temp))
		    (goto-if-true ,loopbottom-label (temp ,finished-temp))

		    ,@(let ((cell-temp (gen-temp "temp"))
			    (stringbuffer-temp (gen-temp "stringbuffer"))
			    (compare-temp (gen-temp "compare"))
			    (skip-label (gen-label "skip")))

			`((declare-temp ,cell-temp)
			  (declare-temp ,stringbuffer-temp)
			  (declare-temp ,compare-temp)

			  (read-array-element (temp ,cell-temp)
					      (temp ,launcher-stackframe-temp)
					      (const ,launcher-varindex)
					      (const 1)
					      (temp ,currentrow-temp)
					      (const 0))

			  (assign (temp ,stringbuffer-temp) (const ,launcher-action))
			  (append (temp ,stringbuffer-temp) (temp ,currentrow-temp))
			  ,@(if (or (null? displaystyle)
				    (eq? (car displaystyle) 'header)
				    (eq? (car displaystyle) 'main)
				    (eq? (car displaystyle) 'footer))

				`((number-equal (temp ,compare-temp) (temp ,currentrow-temp) (const 3))
				  (goto-if-true ,skip-label (temp ,compare-temp))
				  (html ,displaystyle (const "                        | "))
				  ,skip-label
				  (html-launcher-button ,displaystyle (temp ,stringbuffer-temp) (temp ,cell-temp))
				  (html-newline ,displaystyle))

				`((html ,displaystyle (const "                        <p>"))
				  (html-launcher-button ,displaystyle (temp ,stringbuffer-temp) (temp ,cell-temp))
				  (html ,displaystyle (const "</p>"))
				  (html-newline ,displaystyle)))))
		    
		    (increment (temp ,currentrow-temp) (const 1))
		    (goto ,looptop-label)

		    ,loopbottom-label

		    ,@(if (or (null? displaystyle)
			      (eq? (car displaystyle) 'header)
			      (eq? (car displaystyle) 'main)
			      (eq? (car displaystyle) 'footer))
			  `((html ,displaystyle (const "                        ]"))
			    (html-newline ,displaystyle)
			    (html ,displaystyle (const "                        </p>"))
			    (html-newline ,displaystyle))
			  `())

		    (goto ,done-label)
		    
		    ,noentries-label

		    (html ,displaystyle (const "                        <p>Invalid launcher - not initialized.</p>"))
		    (html-newline ,displaystyle)

		    ,done-label

		    )))

	    `((comment "Check whether any of the launcher choices were selected")

	      ,@(let ((stringequal-temp         (gen-temp "stringequal"))
		      (launcher-stackframe-temp (gen-temp "stackframe"))
		      (launcher-option          (gen-temp "option"))
		      (mission-id               (gen-temp "missionid"))
		      (continue-label           (gen-label "continue")))

		  `((declare-temp ,stringequal-temp)
		    (declare-temp ,launcher-stackframe-temp)
		    (declare-temp ,launcher-option)
		    (declare-temp ,mission-id)
		    
		    (string-match (temp ,stringequal-temp) (field control) (const ,launcher-action))
		    (not (temp ,stringequal-temp) (temp ,stringequal-temp))
		    (goto-if-true ,continue-label (temp ,stringequal-temp))

		    ;; One of the options from this launcher was selected.
		    ;; Store the option number in the appropriate slot in the launcher.
		    ;; We've come back from a display, so we need to recompute the stack frame, 
		    ;; but we don't need to redeclare the temporary variable:

		    (comment "Launcher choice selected; save the index number and jump to the launch processing code")

		    ,(if (sym-static-statesafe-launcher? sym)
			 `(assign (temp ,launcher-stackframe-temp) (const 0))
			 `(stack-maxindex (temp ,launcher-stackframe-temp)))

		    (substring (temp ,launcher-option)
			       (field control)
			       (const ,(string-length launcher-action)))

		    (write-array-element (temp ,launcher-option)
					 (temp ,launcher-stackframe-temp)
					 (const ,launcher-varindex)
					 (const 1)
					 (const 1)
					 (const 0))

		    ;; Mission id

		    (read-array-element (temp ,mission-id)
					(temp ,launcher-stackframe-temp)
					(const ,launcher-varindex)
					(const 1)
					(const 2)
					(const 0))
		    
		    (goto-launch (temp ,mission-id))

		    ,continue-label))))))

(define (compile-expr-print-statesafe-text error-exit linenumber displaystyle state active-symtab full-symtab sym target)
  (let ((text-varindex (sym-statesafe-index sym))
	(text-stackframe-temp (gen-temp "stackframe"))
	(text-value-temp (gen-temp (symbol->string (sym-name sym)))))

    `((declare-temp ,text-stackframe-temp)
      (declare-temp ,text-value-temp)

      ,(if (sym-static-statesafe-text? sym)
	   `(comment ,(string-append "Print statesafe text \"" (symbol->string (sym-name sym)) "\" (global; use stack frame 0):"))
	   `(comment ,(string-append "Print statesafe text \"" (symbol->string (sym-name sym)) "\" (in subroutine; use top of stack):")))

      ;; store the text's stack frame in temporary variable

      ,(if (sym-static-statesafe-text? sym)
	   `(assign (temp ,text-stackframe-temp) (const 0))
	   `(stack-maxindex (temp ,text-stackframe-temp)))

      (read-scalar (temp ,text-value-temp) (temp ,text-stackframe-temp) (const ,text-varindex))
      (html ,displaystyle (const "                        <pre>"))
      (html ,displaystyle (temp ,text-value-temp))
      (html ,displaystyle (const "</pre>"))
      (html-newline ,displaystyle))))

(define (compile-expr-print-transient-var error-exit linenumber displaystyle state active-symtab full-symtab sym target)
  `((comment "Print transient variable:")
    (html ,displaystyle (temp ,(sym-transient-varname sym)))))

(define (compile-expr-print-application error-exit linenumber displaystyle state active-symtab full-symtab expr target)
  (let ((temp (gen-temp "result")))
    (let ((instructions (compile-expr-general error-exit linenumber state active-symtab full-symtab expr `(temp ,temp))))
      `((comment "Print general expression:")
	(declare-temp ,temp)
	,@instructions
	(html ,displaystyle (temp ,temp))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-image:
;;;----------------------------------------------------------------------

(define (compile-stmt-image error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(displaystyle (caddr stmt)))

    (if (not (= (length stmt) 8))
	(compile-error error-exit linenumber "image statement must have exactly 4 arguments - src, alt, width, and height"))
    
    (if (not (state-in-display? state))
	(compile-error error-exit linenumber "cannot use image statement outside display"))

    (let ((src (cadr (cdddr stmt)))
	  (alt (caddr (cdddr stmt)))
	  (width (cadddr (cdddr stmt)))
	  (height (cadr (cdddr (cdddr stmt)))))

      (if (not (string? src))
	  (compile-error error-exit linenumber "invalid image src argument - must be literal string"))
      
      (if (not (string? alt))
	  (compile-error error-exit linenumber "invalid image alt argument - must be literal string"))
      
      (if (not (number? width))
	  (compile-error error-exit linenumber "invalid image width argument - must be literal number"))
      
      (if (not (number? height))
	  (compile-error error-exit linenumber "invalid image height argument - must be literal number"))

      (let ((style-id (cond ((null? displaystyle) #f)
			    ((null? (cadr displaystyle)) #f)
			    (else (cadr displaystyle)))))

	(values state
		active-symtab
		full-symtab
		`(,(if style-id
		       `(html ,displaystyle (const ,(string-append "                    <tr class=\"" style-id "\">")))
		       `(html ,displaystyle (const "                    <tr>")))
		  (html-newline ,displaystyle)
		  (html ,displaystyle (const "                      <td class=\"mainDisplayCell\" colspan=\"2\">"))
		  (html-newline ,displaystyle)
		  (html ,displaystyle (const ,(string-append "                        <img src=\"" src
							     "\" alt=\"" alt
							     "\" width=\"" (number->string width)
							     "\" height=\"" (number->string height)
							     "\" />")))
		  (html-newline ,displaystyle)
		  (html ,displaystyle (const "                      </td>"))
		  (html-newline ,displaystyle)
		  (html ,displaystyle (const "                    </tr>"))
		  (html-newline ,displaystyle))
		'()
		'()
		'())))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-edit:
;;;----------------------------------------------------------------------

(define (compile-stmt-edit error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(displaystyle (caddr stmt)))

    (if (not (= (length stmt) 6))
	(compile-error error-exit linenumber "edit statement must have exactly 2 arguments - label and variable"))
    
    (if (not (state-in-display? state))
	(compile-error error-exit linenumber "cannot use edit statement outside display"))

    (if (or (state-in-loop? state) (state-in-conditional? state))
	(compile-error error-exit linenumber "cannot use edit statement inside while loop or conditional statement"))

    (let ((label (cadr (cdddr stmt)))
	  (var (caddr (cdddr stmt))))

      (if (not (string? label))
	  (compile-error error-exit linenumber "invalid edit label - must be literal string"))

      (cond ((expr-variable? var)
	     (let ((sym (search-symtab active-symtab var)))

	       (cond ((not sym)
		      (compile-error error-exit linenumber (string-append "unknown variable \"" (symbol->string var) "\"")))

		     ((or (sym-static-statesafe-var? sym) (sym-auto-statesafe-var? sym))
		      (compile-stmt-edit-statesafe-var error-exit state active-symtab full-symtab displaystyle label sym))

		     ((or (sym-static-statesafe-table? sym) (sym-auto-statesafe-table? sym))
		      (compile-stmt-edit-statesafe-table error-exit state active-symtab full-symtab displaystyle label sym))

		     ((or (sym-static-statesafe-enum? sym) (sym-auto-statesafe-enum? sym))
		      (compile-stmt-edit-statesafe-enum error-exit state active-symtab full-symtab displaystyle label sym))

		     ((or (sym-static-statesafe-text? sym) (sym-auto-statesafe-text? sym))
		      (compile-stmt-edit-statesafe-text error-exit state active-symtab full-symtab displaystyle label sym))

		     ((sym-static-transient-var? sym)
		      (compile-error error-exit linenumber "cannot edit transient variables"))

		     (else
		      (compile-error error-exit linenumber (string-append "variable \"" (symbol->string var) "\" has invalid or unimplemented type for editing"))))))

	    (else
	     (compile-error error-exit linenumber "invalid edit variable"))))))

(define (compile-stmt-edit-statesafe-var error-exit state active-symtab full-symtab displaystyle label sym)

  (let ((inputfield-name (gen-temp (string-append (symbol->string (sym-name sym)) "_field")))
	(var-varindex (sym-statesafe-index sym))
	(var-name (sym-name sym))
	(style-id (cond ((null? displaystyle) #f)
			((null? (cadr displaystyle)) #f)
			(else (cadr displaystyle)))))
    (values state
	    active-symtab
	    full-symtab

	    ;; 4th return value is the list of instructions to be inserted inline into the
	    ;; instruction stream (in the case of the edit statement, this is always within a
 	    ;; display statement).
	    
	    (let ((var-stackframe-temp (gen-temp "stackframe")))
	      `((declare-temp ,var-stackframe-temp)

		,(if (sym-static-statesafe-var? sym)
		     `(comment ,(string-append "Edit statesafe var \"" (symbol->string var-name) "\" (global; use stack frame 0):"))
		     `(comment ,(string-append "Edit statesafe var \"" (symbol->string var-name) "\" (in subroutine; use top of stack):")))

		;; Store the var's stack frame in temporary variable for use throughout this
		;; "edit var" subroutine.

		,(if (sym-static-statesafe-var? sym)
		     `(assign (temp ,var-stackframe-temp) (const 0))
		     `(stack-maxindex (temp ,var-stackframe-temp)))

		;; Generate code to fetch the current stored value.

		,@(let ((value-temp (gen-temp (symbol->string var-name))))
		    `((declare-temp ,value-temp)
		      (read-scalar (temp ,value-temp) (temp ,var-stackframe-temp) (const ,var-varindex))

		      ;; Display the label and a textbox containing the current value.

		      ,(if style-id
			   `(html ,displaystyle (const ,(string-append "                    <tr class=\"" style-id "\">")))
			   `(html ,displaystyle (const "                    <tr>")))

		      (html-newline ,displaystyle)
		      (html ,displaystyle (const "                      <td class=\"labelDisplayCell\"><p>"))
		      (html ,displaystyle (const ,label))
		      (html ,displaystyle (const "</p></td>"))
		      (html-newline ,displaystyle)
		      (html ,displaystyle (const "                      <td class=\"mainDisplayCell\">"))
		      (html-field ,displaystyle ,inputfield-name (temp ,value-temp))
		      (html ,displaystyle (const "</td>"))
		      (html-newline ,displaystyle)
		      (html ,displaystyle (const "                    </tr>"))
		      (html-newline ,displaystyle)))))

	    ;; 5th return value is the list of instructions to be inserted after the current
	    ;; display statement to save statesafe variables whose values may have changed as a
	    ;; result of the form submission.

	    (let ((var-stackframe-temp (gen-temp "stackframe")))
	      `((declare-temp ,var-stackframe-temp)

		;; Store the var's stack frame in temporary variable.

		,(if (sym-static-statesafe-var? sym)
		     `(assign (temp ,var-stackframe-temp) (const 0))
		     `(stack-maxindex (temp ,var-stackframe-temp)))

		(write-scalar (field ,inputfield-name) (temp ,var-stackframe-temp) (const ,var-varindex))))

	    ;; 6th return value is the list of instructions to be inserted after the current
	    ;; display statement to perform any required control transfers. This is never
	    ;; required when editing a statesafe var.

	    '()

	    ;; 7th return value is the list of instructions corresponding to any additional
	    ;; subroutines that need to be added to the end of the program.  No additional
	    ;; subroutines are required when editing a statesafe var.

	    '())))

(define (compile-stmt-edit-statesafe-table error-exit state active-symtab full-symtab displaystyle label sym)

  (let ((table-varindex (sym-statesafe-index sym))
	(table-fields (sym-statesafe-type-fields sym))

	(style-id (cond ((null? displaystyle) #f)
			((null? (cadr displaystyle)) #f)
			(else (cadr displaystyle))))

	;; Values for the add, modify, and remove buttons:

	(addentry-action (gen-action-selector-value "addentry"))
	(modentry-action (gen-action-selector-value "modentry"))
	(rementry-action (gen-action-selector-value "rementry"))

	;; Labels for the add, modify, and remove subroutines:

	(addentry-subroutine-label (gen-label "addentry_subroutine"))
	(modentry-subroutine-label (gen-label "modentry_subroutine"))
	(rementry-subroutine-label (gen-label "rementry_subroutine"))

	;; Symbol table entries for the add, modify, and remove subroutines:

	(add-entry-subroutine-sym (new-sym-subroutine 'generated-add-entry))
	(mod-entry-subroutine-sym (new-sym-subroutine 'generated-modify-entry))
	(rem-entry-subroutine-sym (new-sym-subroutine 'generated-remove-entry)))

    (set! full-symtab (extend-symtab full-symtab add-entry-subroutine-sym))
    (set! full-symtab (extend-symtab full-symtab mod-entry-subroutine-sym))
    (set! full-symtab (extend-symtab full-symtab rem-entry-subroutine-sym))

    (values state
	    active-symtab
	    full-symtab

	    ;; 4th return value is the list of instructions to be inserted inline into the
	    ;; instruction stream (in the case of the edit statement, this is always within a
	    ;; display statement).

	    (let ((table-stackframe-temp (gen-temp "stackframe")))
	      `((declare-temp ,table-stackframe-temp)

		,(if (sym-static-statesafe-table? sym)
		     `(comment ,(string-append "Edit statesafe table \"" (symbol->string (sym-name sym)) "\" (global; use stack frame 0):"))
		     `(comment ,(string-append "Edit statesafe table \"" (symbol->string (sym-name sym)) "\" (in subroutine; use top of stack):")))

		;; Store the var's stack frame in temporary variable for use throughout this
		;; "edit table" subroutine.

		,(if (sym-static-statesafe-table? sym)
		     `(assign (temp ,table-stackframe-temp) (const 0))
		     `(stack-maxindex (temp ,table-stackframe-temp)))

		,(if style-id
		     `(html ,displaystyle (const ,(string-append "                    <tr class=\"" style-id "\">")))
		     `(html ,displaystyle (const "                    <tr>")))
		
		(html-newline ,displaystyle)
		(html ,displaystyle (const "                      <td class=\"mainDisplayCell\" colspan=\"2\">"))
		(html-newline ,displaystyle)

		;; Display the label.

		(html ,displaystyle (const ,(string-append "                        <p>" label "</p>")))
		(html-newline ,displaystyle)

		;; If the table is empty, print a message to that
		;; effect and display an "Add entry ..." button.

		,@(let ((isempty-temp    (gen-temp  "isempty"))
			(noentries-label (gen-label "noentries"))
			(finished-label  (gen-label "finished")))

		    `((declare-temp ,isempty-temp)
		      (array-empty (temp ,isempty-temp) (temp ,table-stackframe-temp) (const ,table-varindex))
		      (goto-if-true ,noentries-label (temp ,isempty-temp))

		      ;; If we've gotten to this point, the table is not empty.
		      ;; Display the full table editor.

		      ,@(let ((currentrow-temp  (gen-temp  "currentrow"))
			      (rowcount-temp    (gen-temp  "rowcount"))
			      (finished-temp    (gen-temp  "finished"))
			      (looptop-label    (gen-label "looptop"))
			      (loopbottom-label (gen-label "loopbottom")))

			  `((declare-temp ,currentrow-temp)
			    (declare-temp ,rowcount-temp)
			    (declare-temp ,finished-temp)

			    (html ,displaystyle (const ,(string-append "                        <table summary=\"" (symbol->string (sym-name sym)) "\">")))
			    (html-newline ,displaystyle)

			    ;; Print out table header (column labels).
			    
			    (html ,displaystyle (const "                          <tr>"))
			    (html-newline ,displaystyle)
			    
			    ,@(let ((instructions '()))
				(for-each (lambda (field)
					    (set! instructions `(,@instructions
								 (html ,displaystyle (const ,(string-append "                            <td class=\"tableHeaderCell\">&nbsp;" (field-label field) "&nbsp;</td>")))
								 (html-newline ,displaystyle))))
					  table-fields)
				instructions)

			    ;; Add two blanks for "modify" and "remove" columns

			    (html ,displaystyle (const "                            <td class=\"tableHeaderCell\">&nbsp;</td>"))
			    (html-newline ,displaystyle)
			    (html ,displaystyle (const "                            <td class=\"tableHeaderCell\">&nbsp;</td>"))
			    (html-newline ,displaystyle)
			    (html ,displaystyle (const "                          </tr>"))
			    (html-newline ,displaystyle)

			    ;; Loop over all the rows, then within each row, loop over all the
			    ;; columns. Read each value into a temporary variable and print it.

			    (assign (temp ,currentrow-temp) (const 0))

			    ,looptop-label

			    (array-rowcount (temp ,rowcount-temp) (temp ,table-stackframe-temp) (const ,table-varindex) (const ,(fieldlist-colcount table-fields)))
			    (number-equal (temp ,finished-temp) (temp ,currentrow-temp) (temp ,rowcount-temp))
			    (goto-if-true ,loopbottom-label (temp ,finished-temp))

			    (html ,displaystyle (const "                          <tr>"))
			    (html-newline ,displaystyle)

			    ,@(let ((cell-temp (gen-temp "temp"))
				    (stringbuffer-temp (gen-temp  "stringbuffer")))

				`((declare-temp ,cell-temp)
				  (declare-temp ,stringbuffer-temp)
				  
				  ,@(let ((tempcode '()))

				      (do ((i 0 (+ i 1)))
					  ((= i (length table-fields)) tempcode)
					(set! tempcode `(,@tempcode
							 (html ,displaystyle (const "                            <td class=\"tableCell\">&nbsp;"))
							 (read-array-element (temp ,cell-temp)
									     (temp ,table-stackframe-temp)
									     (const ,table-varindex)
									     (const ,(fieldlist-colcount table-fields))
									     (temp ,currentrow-temp)
									     (const ,(+ i 1)))
							 (html ,displaystyle (temp ,cell-temp))
							 (html ,displaystyle (const "&nbsp;</td>"))
							 (html-newline ,displaystyle)))))
				  
				  ;; Read the row identifier from column zero, and create the
				  ;; modify and remove buttons as two extra columns.

				  (read-array-element (temp ,cell-temp)
						      (temp ,table-stackframe-temp)
						      (const ,table-varindex)
						      (const ,(fieldlist-colcount table-fields))
						      (temp ,currentrow-temp)
						      (const 0))
				  
				  (assign (temp ,stringbuffer-temp) (const ,modentry-action))
				  (append (temp ,stringbuffer-temp) (temp ,cell-temp))
				  (html ,displaystyle (const "                            <td class=\"tableCell\">"))
				  (html-table-button ,displaystyle (temp ,stringbuffer-temp) (const "Modify ..."))
				  (html ,displaystyle (const "</td>"))
				  (html-newline ,displaystyle)

				  (assign (temp ,stringbuffer-temp) (const ,rementry-action))
				  (append (temp ,stringbuffer-temp) (temp ,cell-temp))
				  (html ,displaystyle (const "                            <td class=\"tableCell\">"))
				  (html-table-button ,displaystyle (temp ,stringbuffer-temp) (const "Remove ..."))
				  (html ,displaystyle (const "</td>"))
				  (html-newline ,displaystyle)))

			    (html ,displaystyle (const "                          </tr>"))
			    (html-newline ,displaystyle)

			    (increment (temp ,currentrow-temp) (const 1))
			    (goto ,looptop-label)

			    ,loopbottom-label

			    ;; Finished printing all the rows. Now display the "add entry"
			    ;; button on the bottom.

			    (html ,displaystyle (const "                          <tr>"))
			    (html-newline ,displaystyle)
			    (html ,displaystyle (const ,(string-append "                            <td class=\"tableAddEntryCell\" colspan=\""
								       (number->string (+ (fieldlist-colcount table-fields) 1))
								       "\">")))
			    (html-table-button ,displaystyle (const ,addentry-action) (const "Add entry ..."))
			    (html ,displaystyle (const "</td>"))
			    (html-newline ,displaystyle)
			    (html ,displaystyle (const "                          </tr>"))
			    (html-newline ,displaystyle)
			    (html ,displaystyle (const "                        </table>"))
			    (html-newline ,displaystyle)))

		      (goto ,finished-label)
		      
		      ,noentries-label

		      ;; If we've gotten to this point, the table is empty.
		      ;; Print a message to that effect and an "Add entry ..." button.

		      (html ,displaystyle (const "                        <p>No entries.</p>"))
		      (html-newline ,displaystyle)

		      (html ,displaystyle (const "                        <p>"))
		      (html-button ,displaystyle (const ,addentry-action) (const "Add entry ..."))
		      (html ,displaystyle (const "</p>"))
		      (html-newline ,displaystyle)

		      ,finished-label

		      (html ,displaystyle (const "                      </td>"))
		      (html-newline ,displaystyle)
		      (html ,displaystyle (const "                    </tr>"))
		      (html-newline ,displaystyle)))))

	    ;; 5th return value is the list of instructions to be inserted after the current
	    ;; display statement to save statesafe variables whose values may have changed as a
	    ;; result of the form submission. This is not required when editing a statesafe table.

	    '()

	    ;; 6th return value is the list of instructions to be inserted after the current
	    ;; display statement to perform any required control transfers. For a statesafe
	    ;; table, we need to check whether the user clicked add, modify, or remove.

	    (let ((stringequal-temp    (gen-temp  "stringequal"))
		  (discard-temp        (gen-temp  "discard"))

		  (do-addentry-label   (gen-label "do_addentry"))
		  (do-modentry-label   (gen-label "do_modentry"))
		  (do-rementry-label   (gen-label "do_rementry"))

		  (dont-addentry-label (gen-label "dont_addentry"))
		  (dont-modentry-label (gen-label "dont_modentry"))
		  (dont-rementry-label (gen-label "dont_rementry"))

		  (current-display-id  (state-display-id state)))

	      `((declare-temp ,stringequal-temp)
		(declare-temp ,discard-temp)
		
		(string-equal (temp ,stringequal-temp) (field control) (const ,addentry-action))
		(goto-if-true ,do-addentry-label (temp ,stringequal-temp))
		(goto ,dont-addentry-label)
		
		,do-addentry-label
		(gosub (temp ,discard-temp)
		       ,(sym-name add-entry-subroutine-sym)
		       ,(sym-subroutine-id add-entry-subroutine-sym)
		       ,addentry-subroutine-label
		       ,(gen-call))
		(goto ,(string-append "top_of_display_" (number->string current-display-id)))

		,dont-addentry-label
		(string-match (temp ,stringequal-temp) (field control) (const ,modentry-action))
		(goto-if-true ,do-modentry-label (temp ,stringequal-temp))
		(goto ,dont-modentry-label)

		,do-modentry-label
		(gosub (temp ,discard-temp)
		       ,(sym-name mod-entry-subroutine-sym)
		       ,(sym-subroutine-id mod-entry-subroutine-sym)
		       ,modentry-subroutine-label
		       ,(gen-call)
		       (field control))
		(goto ,(string-append "top_of_display_" (number->string current-display-id)))

		,dont-modentry-label
		(string-match (temp ,stringequal-temp) (field control) (const ,rementry-action))
		(goto-if-true ,do-rementry-label (temp ,stringequal-temp))
		(goto ,dont-rementry-label)

		,do-rementry-label
		(gosub (temp ,discard-temp)
		       ,(sym-name rem-entry-subroutine-sym)
		       ,(sym-subroutine-id rem-entry-subroutine-sym)
		       ,rementry-subroutine-label
		       ,(gen-call)
		       (field control))
		(goto ,(string-append "top_of_display_" (number->string current-display-id)))

		,dont-rementry-label
		))

	    ;; 7th return value is the list of instructions corresponding to any additional
	    ;; subroutines that need to be added to the end of the program.  For a statesafe
	    ;; table we need subroutines for adding, modifying, and removing records.

	    `((comment ,(string-append "Subroutine to add a record to statesafe table \"" (symbol->string (sym-name sym)) "\":"))

	      ,addentry-subroutine-label

	      ;; Display "add record" form:

	      (display ,(gen-display-id) "add record" no_continue

		       ,(if style-id
			    `(html () (const ,(string-append "                    <tr class=\"" style-id "\">")))
			    `(html () (const "                    <tr>")))

		       (html-newline ())
		       (html () (const "                      <td class=\"mainDisplayCell\" colspan=\"2\">"))
		       (html-newline ())
		       (html () (const "                        <p>Add record:</p>"))
		       (html-newline ())
		       (html () (const ,(string-append "                        <table summary=\"add record (" (symbol->string (sym-name sym)) ")\">")))
		       (html-newline ())

		       ,@(let ((tempcode '()))
			   (do ((i 0 (+ i 1)))
			       ((= i (length table-fields)) tempcode)
			     (set! tempcode `(,@tempcode
					      (html () (const "                          <tr>"))
					      (html-newline ())
					      (html () (const ,(string-append "                            <td class=\"tableCell\">"
									      (field-label (list-ref table-fields i))
									      "</td>")))
					      (html-newline ())
					      (html () (const "                            <td class=\"tableCell\">"))
					      (html-field () ,(field-name (list-ref table-fields i)) (const ""))
					      (html () (const "</td>"))
					      (html-newline ())
					      (html () (const "                          </tr>"))
					      (html-newline ())))))
		       
		       (html () (const "                        </table>"))
		       (html-newline ())
		       (html () (const "                        <p>"))
		       (html-newline ())
		       (html () (const "                          "))
		       (html-button () (const "save") (const "Save"))
		       (html-newline ())
		       (html () (const "                          "))
		       (html-button () (const "cancel") (const "Cancel"))
		       (html-newline ())
		       (html () (const "                        </p>"))
		       (html-newline ())
		       (html () (const "                      </td>"))
		       (html-newline ())
		       (html () (const "                    </tr>"))
		       (html-newline ()))

	      ;; The user has submitted the form to add a new row to this table. To store the
	      ;; data, we need the table's current row count.  (Note that we're in a
	      ;; subroutine called by the "edit table" command - so if we're editing a
	      ;; variable with "static" storage class, we'll find the table variable in stack
	      ;; frame 0, otherwise it's the stack frame 1 below the top.)

	      ,@(let ((table-stackframe-temp     (gen-temp "stackframe"))
		      (rowcount-temp             (gen-temp "rowcount"))
		      (new-rowid-temp            (gen-temp "new_rowid"))
		      (addentry-confirmed-temp   (gen-temp "addentry_confirmed"))
		      (do-addentry-confirm-label (gen-label "do_addentry_confirm"))
		      (do-addentry-cancel-label  (gen-label "do_addentry_cancel")))

		  `((declare-temp ,table-stackframe-temp)
		    (declare-temp ,rowcount-temp)
		    (declare-temp ,new-rowid-temp)
		    (declare-temp ,addentry-confirmed-temp)

		    (string-equal (temp ,addentry-confirmed-temp) (field control) (const "save"))
		    (goto-if-true ,do-addentry-confirm-label (temp ,addentry-confirmed-temp))
		    (goto ,do-addentry-cancel-label)

		    ,do-addentry-confirm-label
		    
		    ,@(if (sym-static-statesafe-table? sym)
			  `((assign (temp ,table-stackframe-temp) (const 0)))
			  `((stack-maxindex (temp ,table-stackframe-temp))
			    (decrement (temp ,table-stackframe-temp) (const 1))))

		    (array-rowcount (temp ,rowcount-temp)
				    (temp ,table-stackframe-temp)
				    (const ,table-varindex)
				    (const ,(fieldlist-colcount table-fields)))

		    ;; Generate a new row identifier and write it to the zeroth column of the
		    ;; new row.

		    (generate-rowid (temp ,new-rowid-temp))
		    (write-array-element (temp ,new-rowid-temp)
					 (temp ,table-stackframe-temp)
					 (const ,table-varindex)
					 (const ,(fieldlist-colcount table-fields))
					 (temp ,rowcount-temp)
					 (const 0))

		    ;; Copy the values from each field into the remaining columns of the new
		    ;; row.

		    ,@(let ((tempcode '()))
			(do ((i 0 (+ i 1)))
			    ((= i (length table-fields)) tempcode)
			  (set! tempcode `(,@tempcode
					   (write-array-element (field ,(field-name (list-ref table-fields i)))
								(temp ,table-stackframe-temp)
								(const ,table-varindex)
								(const ,(fieldlist-colcount table-fields))
								(temp ,rowcount-temp)
								(const ,(+ i 1)))))))

		    ,do-addentry-cancel-label

		    ;; Return to caller.

		    (return)))

	      (comment ,(string-append "Subroutine to modify a record in a statesafe table \"" (symbol->string (sym-name sym)) "\":"))

	      ,modentry-subroutine-label
	      
	      ;; Display "modify record" form:

	      (display ,(gen-display-id) "modify record" no_continue

		       ,(if style-id
			    `(html () (const ,(string-append "                    <tr class=\"" style-id "\">")))
			    `(html () (const "                    <tr>")))

		       (html-newline ())
		       (html () (const "                      <td class=\"mainDisplayCell\" colspan=\"2\">"))
		       (html-newline ())
		       (html () (const "                        <p>Modify record:</p>"))
		       (html-newline ())

		       ;; We are modifying a row in this table. We need to read the existing
		       ;; data in the row so we can display it for editing.  The row is
		       ;; identified by the "action" command passed as the first argument to
		       ;; this subroutine, so we will find it in slot 3 on the top stack
		       ;; frame. (First subroutine arg is always in slot 3; slots 0-2 are
		       ;; reserved by the system.)

		       ,@(let ((table-stackframe-temp      (gen-temp "stackframe"))
			       (subroutine-stackframe-temp (gen-temp "subroutine_stackframe"))
			       (modentry-action-rowid-temp (gen-temp "modentry_action_rowid"))
			       (modentry-rowid-temp 	   (gen-temp "modentry_rowid"))
			       (modentry-rowindex-temp     (gen-temp "modentry_rowindex")))

			   `((declare-temp ,table-stackframe-temp)
			     (declare-temp ,subroutine-stackframe-temp)
			     (declare-temp ,modentry-action-rowid-temp)
			     (declare-temp ,modentry-rowid-temp)
			     (declare-temp ,modentry-rowindex-temp)

			     (stack-maxindex (temp ,subroutine-stackframe-temp))
			     (read-scalar (temp ,modentry-action-rowid-temp) (temp ,subroutine-stackframe-temp) (const 3))

			     ;; Now we have the "action" command, which is in the form:
			     ;; 
			     ;;   modentry-M-NNNNNNNN
			     ;; 
			     ;; where M is a counter, and NNNNNNNN is the id of the row we want to edit.
			     ;; Next, use the substring command to remove the prefix, leaving only the id:

			     (substring (temp ,modentry-rowid-temp)
					(temp ,modentry-action-rowid-temp)
					(const ,(string-length modentry-action)))

			     ;; Get the stackframe of the table we are trying to edit.
			     ;; Note that we're in a subroutine called by the "edit table"
			     ;; command - so if we're editing a variable with "auto" storage
			     ;; class, we'll find the variable in the stack frame one below
			     ;; the top. Otherwise the stack frame is 0.

			     ,@(if (sym-static-statesafe-table? sym)
				   `((assign (temp ,table-stackframe-temp) (const 0)))
				   `((stack-maxindex (temp ,table-stackframe-temp))
				     (decrement (temp ,table-stackframe-temp) (const 1))))

			     ;; Search the table for the given row id to get a row index
			     ;; (i.e. the actual row number)

			     (lookup-array-index (temp ,modentry-rowindex-temp)
						 (temp ,modentry-rowid-temp)
						 (temp ,table-stackframe-temp)
						 (const ,table-varindex)
						 (const ,(fieldlist-colcount table-fields)))

			     ;; Finally we can display the values to edit:

			     (html () (const ,(string-append "                        <table summary=\"modify record (" (symbol->string (sym-name sym)) ")\">")))
			     (html-newline ())

			     ,@(let ((cell-temp (gen-temp "temp")))
				 (let ((tempcode `((declare-temp ,cell-temp))))

				   (do ((i 0 (+ i 1)))
				       ((= i (length table-fields)) tempcode)
				     (set! tempcode `(,@tempcode
						      (html () (const "                          <tr>"))
						      (html-newline ())
						      (html () (const ,(string-append "                            <td class=\"tableCell\">"
										      (field-label (list-ref table-fields i))
										      "</td>")))
						      (html-newline ())
						      (html () (const "                            <td class=\"tableCell\">"))
						      (read-array-element (temp ,cell-temp)
									  (temp ,table-stackframe-temp)
									  (const ,table-varindex)
									  (const ,(fieldlist-colcount table-fields))
									  (temp ,modentry-rowindex-temp)
									  (const ,(+ i 1)))
						      (html-field () ,(field-name (list-ref table-fields i)) (temp ,cell-temp))
						      (html () (const "</td>"))
						      (html-newline ())
						      (html () (const "                          </tr>"))
						      (html-newline ()))))))

			     (html () (const "                        </table>"))
			     (html-newline ())
			     (html () (const "                        <p>"))
			     (html-newline ())
			     (html () (const "                          "))
			     (html-button () (const "save") (const "Save"))
			     (html-newline ())
			     (html () (const "                          "))
			     (html-button () (const "cancel") (const "Cancel"))
			     (html-newline ())
			     (html () (const "                        </p>"))
			     (html-newline ())
			     (html () (const "                      </td>"))
			     (html-newline ())
			     (html () (const "                    </tr>"))
			     (html-newline ()))))

	      ;; We are modifying a row in this table. The user has entered the new row
	      ;; values. Check whether answer is "save" or "cancel" and take the appropriate
	      ;; action.

	      ,@(let ((table-stackframe-temp      (gen-temp  "stackframe"))
		      (subroutine-stackframe-temp (gen-temp  "subroutine_stackframe"))
		      (modentry-confirmed-temp    (gen-temp  "modentry_confirmed"))
		      (modentry-action-rowid-temp (gen-temp  "modentry_action_rowid"))
		      (modentry-rowid-temp 	  (gen-temp  "modentry_rowid"))
		      (modentry-rowindex-temp     (gen-temp  "modentry_rowindex"))			
		      (do-modentry-confirm-label  (gen-label "do_modentry_confirm"))
		      (do-modentry-cancel-label   (gen-label "do_modentry_cancel")))

		  `((declare-temp ,table-stackframe-temp)
		    (declare-temp ,subroutine-stackframe-temp)
		    (declare-temp ,modentry-confirmed-temp)
		    (declare-temp ,modentry-action-rowid-temp)
		    (declare-temp ,modentry-rowid-temp)
		    (declare-temp ,modentry-rowindex-temp)

		    (string-equal (temp ,modentry-confirmed-temp) (field control) (const "save"))
		    (goto-if-true ,do-modentry-confirm-label (temp ,modentry-confirmed-temp))
		    (goto ,do-modentry-cancel-label)

		    ,do-modentry-confirm-label

		    ;; User has confirmed the modification.  Get the "action command"
		    ;; subroutine argument (slot 3 on the top stack frame; remember that slots
		    ;; 0-2 are reserved by the system). From that, figure out the index of the
		    ;; table row that needs to be updated.  Then copy the new values into the
		    ;; table row.

		    (stack-maxindex (temp ,subroutine-stackframe-temp))
		    (read-scalar (temp ,modentry-action-rowid-temp) (temp ,subroutine-stackframe-temp) (const 3))

		    ;; Now we have the "action" command, which is in the form:
		    ;; 
		    ;;   modentry-M-NNNNNNNN
		    ;; 
		    ;; where M is a counter, and NNNNNNNN is the id of the row we want to edit.
		    ;; Next, use the substring command to remove the prefix, leaving only the id:

		    (substring (temp ,modentry-rowid-temp)
			       (temp ,modentry-action-rowid-temp)
			       (const ,(string-length modentry-action)))

		    ;; Get the stackframe of the table we are trying to edit.
		    ;; Note that we're in a subroutine called by the "edit table"
		    ;; command - so if we're editing a variable with "auto" storage
		    ;; class, we'll find the variable in the stack frame one below
		    ;; the top. Otherwise the stack frame is 0.

		    ,@(if (sym-static-statesafe-table? sym)
			  `((assign (temp ,table-stackframe-temp) (const 0)))
			  `((stack-maxindex (temp ,table-stackframe-temp))
			    (decrement (temp ,table-stackframe-temp) (const 1))))

		    ;; Search the table for the given row id to get a row index
		    ;; (i.e. the actual row number)

		    (lookup-array-index (temp ,modentry-rowindex-temp)
					(temp ,modentry-rowid-temp)
					(temp ,table-stackframe-temp)
					(const ,table-varindex)
					(const ,(fieldlist-colcount table-fields)))

		    ;; Finally we can store the modified values:

		    ,@(let ((tempcode `()))
			(do ((i 0 (+ i 1)))
			    ((= i (length table-fields)) tempcode)
			  (set! tempcode `(,@tempcode
					   (write-array-element (field ,(field-name (list-ref table-fields i)))
								(temp ,table-stackframe-temp)
								(const ,table-varindex)
								(const ,(fieldlist-colcount table-fields))
								(temp ,modentry-rowindex-temp)
								(const ,(+ i 1)))))))

		    ,do-modentry-cancel-label

		    ;; Return to caller.

		    (return)))

	      (comment ,(string-append "Subroutine to remove a record from a statesafe table \"" (symbol->string (sym-name sym)) "\":"))

	      ,rementry-subroutine-label							      
	      
	      ;; "action command" argument is safely stashed away on top stack frame. Ask user
	      ;; to confirm the deletion.

	      (display ,(gen-display-id) "remove record" no_continue

		       ,(if style-id
			    `(html () (const ,(string-append "                    <tr class=\"" style-id "\">")))
			    `(html () (const "                    <tr>")))

		       (html-newline ())
		       (html () (const "                      <td class=\"mainDisplayCell\" colspan=\"2\">"))
		       (html-newline ())
		       (html () (const "                        <p>Remove record:</p>"))
		       (html-newline ())

		       ;; We are deleting a row in this table. We need to read the existing
		       ;; data in the row so we can display it for confirmation.  The row is
		       ;; identified by the "action" command passed as the first argument to
		       ;; this subroutine, so we will find it in slot 3 on the top stack
		       ;; frame. (First subroutine arg is always in slot 3; slots 0-2 are
		       ;; reserved by the system.)

		       ,@(let ((table-stackframe-temp      (gen-temp "stackframe"))
			       (subroutine-stackframe-temp (gen-temp "subroutine_stackframe"))
			       (rementry-action-rowid-temp (gen-temp "rementry_action_rowid"))
			       (rementry-rowid-temp 	   (gen-temp "rementry_rowid"))
			       (rementry-rowindex-temp     (gen-temp "rementry_rowindex")))

			   `((declare-temp ,table-stackframe-temp)
			     (declare-temp ,subroutine-stackframe-temp)
			     (declare-temp ,rementry-action-rowid-temp)
			     (declare-temp ,rementry-rowid-temp)
			     (declare-temp ,rementry-rowindex-temp)

			     (stack-maxindex (temp ,subroutine-stackframe-temp))
			     (read-scalar (temp ,rementry-action-rowid-temp) (temp ,subroutine-stackframe-temp) (const 3))

			     ;; Now we have the "action" command, which is in the form:
			     ;; 
			     ;;   rementry-M-NNNNNNNN
			     ;; 
			     ;; where M is a counter, and NNNNNNNN is the id of the row we want to edit.
			     ;; Next, use the substring command to remove the prefix, leaving only the id:

			     (substring (temp ,rementry-rowid-temp)
					(temp ,rementry-action-rowid-temp)
					(const ,(string-length rementry-action)))

			     ;; Get the stackframe of the table we are trying to edit.
			     ;; Note that we're in a subroutine called by the "edit table"
			     ;; command - so if we're editing a variable with "auto" storage
			     ;; class, we'll find the variable in the stack frame one below
			     ;; the top. Otherwise the stack frame is 0.

			     ,@(if (sym-static-statesafe-table? sym)
				   `((assign (temp ,table-stackframe-temp) (const 0)))
				   `((stack-maxindex (temp ,table-stackframe-temp))
				     (decrement (temp ,table-stackframe-temp) (const 1))))

			     ;; Search the table for the given row id to get a row index
			     ;; (i.e. the actual row number)

			     (lookup-array-index (temp ,rementry-rowindex-temp)
						 (temp ,rementry-rowid-temp)
						 (temp ,table-stackframe-temp)
						 (const ,table-varindex)
						 (const ,(fieldlist-colcount table-fields)))

			     ;; Finally we can display the values to remove:

			     (html () (const ,(string-append "                        <table summary=\"remove record (" (symbol->string (sym-name sym)) ")\">")))
			     (html-newline ())

			     ,@(let ((cell-temp (gen-temp "temp")))
				 (let ((tempcode `((declare-temp ,cell-temp))))

				   (do ((i 0 (+ i 1)))
				       ((= i (length table-fields)) tempcode)
				     (set! tempcode `(,@tempcode
						      (html () (const "                          <tr>"))
						      (html-newline ())
						      (html () (const ,(string-append "                            <td class=\"tableCell\">"
										      (field-label (list-ref table-fields i))
										      "</td>")))
						      (html-newline ())
						      (html () (const "                            <td class=\"tableCell\">"))
						      (read-array-element (temp ,cell-temp)
									  (temp ,table-stackframe-temp)
									  (const ,table-varindex)
									  (const ,(fieldlist-colcount table-fields))
									  (temp ,rementry-rowindex-temp)
									  (const ,(+ i 1)))
						      (html () (temp ,cell-temp))
						      (html () (const "</td>"))
						      (html-newline ())
						      (html () (const "                          </tr>"))
						      (html-newline ()))))))

			     (html () (const "                        </table>"))
			     (html-newline ())
			     (html () (const "                        <p>"))
			     (html-newline ())
			     (html () (const "                          "))
			     (html-button () (const "confirm") (const "Confirm"))
			     (html-newline ())
			     (html () (const "                          "))
			     (html-button () (const "cancel") (const "Cancel"))
			     (html-newline ())
			     (html () (const "                        </p>"))
			     (html-newline ())
			     (html () (const "                      </td>"))
			     (html-newline ())
			     (html () (const "                    </tr>"))
			     (html-newline ()))))

	      ;; User has replied. Check whether answer is "confirm" or "cancel" and take the
	      ;; appropriate action.

	      ,@(let ((table-stackframe-temp      (gen-temp  "stackframe"))
		      (subroutine-stackframe-temp (gen-temp  "subroutine_stackframe"))
		      (rementry-confirmed-temp    (gen-temp  "rementry_confirmed"))
		      (rementry-action-rowid-temp (gen-temp  "rementry_action_rowid"))
		      (rementry-rowid-temp 	  (gen-temp  "rementry_rowid"))
		      (rementry-rowindex-temp     (gen-temp  "rementry_rowindex"))
		      (do-rementry-confirm-label  (gen-label "do_rementry_confirm"))
		      (do-rementry-cancel-label   (gen-label "do_rementry_cancel")))

		  `((declare-temp ,table-stackframe-temp)
		    (declare-temp ,subroutine-stackframe-temp)
		    (declare-temp ,rementry-confirmed-temp)
		    (declare-temp ,rementry-action-rowid-temp)
		    (declare-temp ,rementry-rowid-temp)
		    (declare-temp ,rementry-rowindex-temp)

		    (string-equal (temp ,rementry-confirmed-temp) (field control) (const "confirm"))
		    (goto-if-true ,do-rementry-confirm-label (temp ,rementry-confirmed-temp))
		    (goto ,do-rementry-cancel-label)

		    ,do-rementry-confirm-label

		    ;; User has confirmed the deletion.  Get the "action command" subroutine
		    ;; argument (slot 3 on the top stack frame; remember that slots 0-2 are
		    ;; reserved by the system). From that, figure out the index of the table
		    ;; row that needs to be deleted.  Then delete the row.

		    (stack-maxindex (temp ,subroutine-stackframe-temp))
		    (read-scalar (temp ,rementry-action-rowid-temp) (temp ,subroutine-stackframe-temp) (const 3))

		    ;; Now we have the "action" command, which is in the form:
		    ;; 
		    ;;   rementry-M-NNNNNNNN
		    ;; 
		    ;; where M is a counter, and NNNNNNNN is the id of the row we want to remove.
		    ;; Next, use the substring command to remove the prefix, leaving only the id:

		    (substring (temp ,rementry-rowid-temp)
			       (temp ,rementry-action-rowid-temp)
			       (const ,(string-length rementry-action)))

		    ;; Get the stackframe of the table we are trying to edit.

		    ,@(if (sym-static-statesafe-table? sym)
			  `((assign (temp ,table-stackframe-temp) (const 0)))
			  `((stack-maxindex (temp ,table-stackframe-temp))
			    (decrement (temp ,table-stackframe-temp) (const 1))))

		    ;; Search the table for the given row id to get a row index
		    ;; (i.e. the actual row number)

		    (lookup-array-index (temp ,rementry-rowindex-temp)
					(temp ,rementry-rowid-temp)
					(temp ,table-stackframe-temp)
					(const ,table-varindex)
					(const ,(fieldlist-colcount table-fields)))

		    ;; Finally we can delete the row:

		    (delete-array-row (temp ,table-stackframe-temp)
				      (const ,table-varindex)
				      (const ,(fieldlist-colcount table-fields))
				      (temp ,rementry-rowindex-temp))

		    ,do-rementry-cancel-label

		    ;; User has cancelled the deletion. Return immediately.
		    
		    (return)))))))

(define (compile-stmt-edit-statesafe-enum error-exit state active-symtab full-symtab displaystyle label sym)

  (let ((inputfield-name (gen-temp (string-append (symbol->string (sym-name sym)) "_field")))
	(enum-varindex (sym-statesafe-index sym))
	(enum-fields (sym-statesafe-type-fields sym))

	(style-id (cond ((null? displaystyle) #f)
			((null? (cadr displaystyle)) #f)
			(else (cadr displaystyle)))))

    (values state
	    active-symtab
	    full-symtab

	    ;; 4th return value is the list of instructions to be inserted inline into the
	    ;; instruction stream (in the case of the edit statement, this is always within a
	    ;; display statement).

	    (let ((enum-stackframe-temp (gen-temp "stackframe")))
	      `((declare-temp ,enum-stackframe-temp)

		,(if (sym-static-statesafe-enum? sym)
		     `(comment ,(string-append "Edit statesafe enum \"" (symbol->string (sym-name sym)) "\" (global; use stack frame 0):"))
		     `(comment ,(string-append "Edit statesafe enum \"" (symbol->string (sym-name sym)) "\" (in subroutine; use top of stack):")))
	      
		;; Get the stackframe of the enum we are trying to edit.

		,(if (sym-static-statesafe-enum? sym)
		     `(assign (temp ,enum-stackframe-temp) (const 0))
		     `(stack-maxindex (temp ,enum-stackframe-temp)))
	      
		,(if style-id
		     `(html ,displaystyle (const ,(string-append "                    <tr class=\"" style-id "\">")))
		     `(html ,displaystyle (const "                    <tr>")))
		      
		;; Display the label:

		(html-newline ,displaystyle)
		(html ,displaystyle (const "                      <td class=\"labelDisplayCell\"><p>"))
		(html ,displaystyle (const ,label))
		(html ,displaystyle (const "</p></td>"))
		(html-newline ,displaystyle)

		;; Display an HTML "select" input (i.e. drop-down menu) to input the enum value.
		;; Make sure the drop-down menu defaults to the current value (by putting the
		;; word "selected" in the appropriate option tag).

		(html ,displaystyle (const "                      <td class=\"mainDisplayCell\">"))
		(html-newline ,displaystyle)

		,@(let ((enum-value-temp (gen-temp (symbol->string (sym-name sym))))
			(stringequal-temp (gen-temp "stringequal")))

		    `((declare-temp ,enum-value-temp)
		      (declare-temp ,stringequal-temp)

		      (read-scalar (temp ,enum-value-temp) (temp ,enum-stackframe-temp) (const ,enum-varindex))
		      (html ,displaystyle (const ,(string-append "                        <select name=\"" inputfield-name "\">")))
		      (html-newline ,displaystyle)
	      
		      ,@(let ((instructions '()))
			  (for-each (lambda (field)
				      (let ((foundmatch-label (gen-label "foundmatch"))
					    (continue-label (gen-label "continue")))
					(set! instructions `(,@instructions
							     (string-equal (temp ,stringequal-temp) (temp ,enum-value-temp) (const ,(symbol->string (field-name field))))
							     (goto-if-true ,foundmatch-label (temp ,stringequal-temp))
							     (html ,displaystyle (const ,(string-append "                          <option value=\""
											  (symbol->string (field-name field))
											  "\">"
											  (field-label field)
											  "</option>")))
							     (html-newline ,displaystyle)
							     (goto ,continue-label)
							     ,foundmatch-label
							     (html ,displaystyle (const ,(string-append "                          <option value=\""
											  (symbol->string (field-name field))
											  "\" selected=\"selected\">"
											  (field-label field)
											  "</option>")))
							     (html-newline ,displaystyle)
							     ,continue-label))))
				    enum-fields)
			  instructions)

		      (html ,displaystyle (const "                        </select>"))
		      (html-newline ,displaystyle)))
	      
		(html ,displaystyle (const "                      </td>"))
		(html-newline ,displaystyle)
		(html ,displaystyle (const "                    </tr>"))
		(html-newline ,displaystyle)))

	    ;; 5th return value is the list of instructions to be inserted after the current
	    ;; display statement to save statesafe variables whose values may have changed as a
	    ;; result of the form submission.

	    (let ((enum-stackframe-temp (gen-temp "stackframe")))
	      `((declare-temp ,enum-stackframe-temp)

		,(if (sym-static-statesafe-enum? sym)
		     `(assign (temp ,enum-stackframe-temp) (const 0))
		     `(stack-maxindex (temp ,enum-stackframe-temp)))
	      
		(write-scalar (field ,inputfield-name) (temp ,enum-stackframe-temp) (const ,enum-varindex))))

	    ;; 6th return value is the list of instructions to be inserted after the current
	    ;; display statement to perform any required control transfers. None required for
	    ;; a statesafe enum.

	    '()

	    ;; 7th return value is the list of instructions corresponding to any additional
	    ;; subroutines that need to be added to the end of the program.  None required for
	    ;; a statesafe enum.

	    '())))

(define (compile-stmt-edit-statesafe-text error-exit state active-symtab full-symtab displaystyle label sym)

  (let ((inputfield-name (gen-temp (string-append (symbol->string (sym-name sym)) "_field")))
	(var-varindex (sym-statesafe-index sym))
	(var-name (sym-name sym))
	(style-id (cond ((null? displaystyle) #f)
			((null? (cadr displaystyle)) #f)
			(else (cadr displaystyle)))))
    (values state
	    active-symtab
	    full-symtab

	    ;; 4th return value is the list of instructions to be inserted inline into the
	    ;; instruction stream (in the case of the edit statement, this is always within a
 	    ;; display statement).
	    
	    (let ((var-stackframe-temp (gen-temp "stackframe")))
	      `((declare-temp ,var-stackframe-temp)

		,(if (sym-static-statesafe-var? sym)
		     `(comment ,(string-append "Edit statesafe text variable \"" (symbol->string var-name) "\" (global; use stack frame 0):"))
		     `(comment ,(string-append "Edit statesafe text variable \"" (symbol->string var-name) "\" (in subroutine; use top of stack):")))

		;; Store the var's stack frame in temporary variable for use throughout this
		;; "edit var" subroutine.

		,(if (sym-static-statesafe-var? sym)
		     `(assign (temp ,var-stackframe-temp) (const 0))
		     `(stack-maxindex (temp ,var-stackframe-temp)))

		;; Generate code to fetch the current stored value.

		,@(let ((value-temp (gen-temp (symbol->string var-name))))
		    `((declare-temp ,value-temp)
		      (read-scalar (temp ,value-temp) (temp ,var-stackframe-temp) (const ,var-varindex))

		      ;; Display the label and a textbox containing the current value.

		      ,(if style-id
			   `(html ,displaystyle (const ,(string-append "                    <tr class=\"" style-id "\">")))
			   `(html ,displaystyle (const "                    <tr>")))

		      (html-newline ,displaystyle)
		      (html ,displaystyle (const "                      <td class=\"mainDisplayCell\" colspan=\"2\">"))
		      (html-newline ,displaystyle)

		      (html ,displaystyle (const ,(string-append "                        <p>" label "</p>")))
		      (html-newline ,displaystyle)

		      (html ,displaystyle (const "                        <p>"))
		      (html-textarea ,displaystyle ,inputfield-name (temp ,value-temp))
		      (html ,displaystyle (const "</p>"))
		      (html-newline ,displaystyle)

		      (html ,displaystyle (const "                      </td>"))
		      (html-newline ,displaystyle)
		      (html ,displaystyle (const "                    </tr>"))
		      (html-newline ,displaystyle)))))

	    ;; 5th return value is the list of instructions to be inserted after the current
	    ;; display statement to save statesafe variables whose values may have changed as a
	    ;; result of the form submission.

	    (let ((var-stackframe-temp (gen-temp "stackframe")))
	      `((declare-temp ,var-stackframe-temp)

		;; Store the var's stack frame in temporary variable.

		,(if (sym-static-statesafe-var? sym)
		     `(assign (temp ,var-stackframe-temp) (const 0))
		     `(stack-maxindex (temp ,var-stackframe-temp)))

		(write-scalar (field ,inputfield-name) (temp ,var-stackframe-temp) (const ,var-varindex))))

	    ;; 6th return value is the list of instructions to be inserted after the current
	    ;; display statement to perform any required control transfers. This is never
	    ;; required when editing a statesafe var.

	    '()

	    ;; 7th return value is the list of instructions corresponding to any additional
	    ;; subroutines that need to be added to the end of the program.  No additional
	    ;; subroutines are required when editing a statesafe var.

	    '())))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-include:
;;;----------------------------------------------------------------------

(define (compile-stmt-include error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(displaystyle (caddr stmt)))

    (if (not (= (length stmt) 5))
	(compile-error error-exit linenumber "wrong number of arguments for \"include\" statement - expecting exactly 1 argument"))
    
    (if (not (null? displaystyle))
	(compile-error error-exit linenumber "display style not applicable to \"include\" statement"))

    (let ((include-item (cadr (cdddr stmt))))

      (if (not (string? include-item))
	  (compile-error error-exit linenumber "invalid include specification - must be literal string"))
      
      (values state
	      active-symtab
	      full-symtab
	      `((include ,include-item))
	      '()
	      '()
	      '()))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-assign:
;;;----------------------------------------------------------------------

(define (compile-stmt-assign error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(varname (caddr stmt))
	(value (cadddr stmt)))

	;; Handle var:

	(cond ((expr-variable? varname)
	       (let ((sym (search-symtab active-symtab varname)))

		 (cond ((not sym)
			(compile-error error-exit linenumber (string-append "unknown variable \"" (symbol->string varname) "\"")))

		       ;; For assignment, treat an "enum" or a "text" like a "var":

		       ((or (sym-static-statesafe-var? sym)  (sym-auto-statesafe-var? sym)
			    (sym-static-statesafe-enum? sym) (sym-auto-statesafe-enum? sym)
			    (sym-static-statesafe-text? sym) (sym-auto-statesafe-text? sym))
			(if (state-in-display? state)
			    (compile-error error-exit linenumber "cannot assign to statesafe variable inside display")
			    (compile-stmt-assign-statesafe-var error-exit state active-symtab full-symtab sym value linenumber)))

		       ((or (sym-static-transient-var? sym) (sym-auto-transient-var? sym))
			(compile-stmt-assign-transient-var error-exit state active-symtab full-symtab sym value linenumber))

		       (else
			(compile-error error-exit linenumber (string-append "variable \"" (symbol->string varname) "\" has invalid or unimplemented type for assignment"))))))
	      
	      (else
	       (compile-error error-exit linenumber "invalid assignment left hand side")))))

(define (compile-stmt-assign-statesafe-var error-exit state active-symtab full-symtab sym value linenumber)
  (let ((temp (gen-temp "source"))
	(stackframe-temp (gen-temp "stackframe")))
    (let ((instructions (compile-expr-general error-exit linenumber state active-symtab full-symtab value `(temp ,temp))))
      (values state
	      active-symtab
	      full-symtab
	      `((comment ,(string-append "Start of assignment to statesafe var \"" (symbol->string (sym-name sym)) "\":"))
		(declare-temp ,temp)
		,@instructions
		(declare-temp ,stackframe-temp)
		,(if (sym-static-statesafe-var? sym)
		     `(assign (temp ,stackframe-temp) (const 0))
		     `(stack-maxindex (temp ,stackframe-temp)))
		(write-scalar (temp ,temp) (temp ,stackframe-temp) (const ,(sym-statesafe-index sym)))
		(comment ,(string-append "End of assignment to statesafe var \"" (symbol->string (sym-name sym)) "\".")))
	      '()
	      '()
	      '()))))

(define (compile-stmt-assign-transient-var error-exit state active-symtab full-symtab sym value linenumber)
  (let ((instructions (compile-expr-general error-exit linenumber state active-symtab full-symtab value `(temp ,(sym-transient-varname sym)))))
    (values state
	    active-symtab
	    full-symtab
	    `((comment ,(string-append "Start of assignment to transient var \"" (symbol->string (sym-name sym)) "\":"))
	      ,@instructions
	      (comment ,(string-append "End of assignment to transient var \"" (symbol->string (sym-name sym)) "\".")))
	    '()
	    '()
	    '())))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-if:
;;;----------------------------------------------------------------------

(define (compile-stmt-if error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(test (caddr stmt))
	(consequent (cadddr stmt))
	(temp (gen-temp "test"))
	(fi (gen-label "fi")))

    (let ((test-inst (compile-expr-general error-exit linenumber state active-symtab full-symtab test `(temp ,temp))))
      (call-with-values (lambda () (compile-stmtlist error-exit
						     (extend-state state (state-in-conditional))
						     active-symtab
						     full-symtab
						     consequent
						     target))
	(lambda (new-state
		 new-active-symtab
		 new-full-symtab
		 new-instructions
		 new-post-state-instructions
		 new-post-control-instructions
		 new-subroutine-instructions)

	  (values state
		  active-symtab
		  new-full-symtab
		  `((comment "Start of \"if\" statement:")
		    (declare-temp ,temp)
		    ,@test-inst
		    (not (temp ,temp) (temp ,temp))
		    (goto-if-true ,fi (temp ,temp))
		    ,@new-instructions
		    ,fi
		    (comment "End of \"if\" statement."))
		  '()
		  '()
		  new-subroutine-instructions))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-ifelse:
;;;----------------------------------------------------------------------

(define (compile-stmt-ifelse error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(test (caddr stmt))
	(consequent (cadddr stmt))
	(alternative (cadr (cdddr stmt)))
	(temp (gen-temp "test"))
	(elselabel (gen-label "else"))
	(fi (gen-label "fi")))

    (let ((test-inst (compile-expr-general error-exit linenumber state active-symtab full-symtab test `(temp ,temp))))
      (call-with-values (lambda () (compile-stmtlist error-exit
						     (extend-state state (state-in-conditional))
						     active-symtab
						     full-symtab
						     consequent
						     target))
	(lambda (con-state
		 con-active-symtab
		 con-full-symtab
		 con-instructions
		 con-post-state-instructions
		 con-post-control-instructions
		 con-subroutine-instructions)

	  (call-with-values (lambda () (compile-stmtlist error-exit
							 (extend-state state (state-in-conditional))
							 active-symtab
							 con-full-symtab
							 alternative
							 target))

	    (lambda (alt-state
		     alt-active-symtab
		     alt-full-symtab
		     alt-instructions
		     alt-post-state-instructions
		     alt-post-control-instructions
		     alt-subroutine-instructions)
	      
	      (values state
		      active-symtab
		      alt-full-symtab
		      `((comment "Start of \"if-else\" statement:")
			(declare-temp ,temp)
			,@test-inst
			(not (temp ,temp) (temp ,temp))
			(goto-if-true ,elselabel (temp ,temp))
			,@con-instructions
			(goto ,fi)
			,elselabel
			,@alt-instructions
			,fi
			(comment "End of \"if-else\" statement."))
		      '()
		      '()
		      `(,@con-subroutine-instructions ,@alt-subroutine-instructions)))))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-while:
;;;----------------------------------------------------------------------

(define (compile-stmt-while error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber (cadr stmt))
	(test (caddr stmt))
	(body (cadddr stmt))
	(temp (gen-temp "test"))
	(topwhile (gen-label "topwhile"))
	(endwhile (gen-label "endwhile")))

    (let ((test-inst (compile-expr-general error-exit linenumber state active-symtab full-symtab test `(temp ,temp))))
      (call-with-values (lambda () (compile-stmtlist error-exit
						     (extend-state state (state-in-loop endwhile))
						     active-symtab
						     full-symtab
						     body
						     target))
	(lambda (new-state
		 new-active-symtab
		 new-full-symtab
		 new-instructions
		 new-post-state-instructions
		 new-post-control-instructions
		 new-subroutine-instructions)

	  (values state
		  active-symtab
		  new-full-symtab
		  `((comment "Start of \"while\" statement:")
		    (declare-temp ,temp)
		    ,topwhile
		    ,@test-inst
		    (not (temp ,temp) (temp ,temp))
		    (goto-if-true ,endwhile (temp ,temp))
		    ,@new-instructions
		    (goto ,topwhile)
		    ,endwhile
		    (comment "End of \"while\" statement."))
		  '()
		  '()
		  new-subroutine-instructions))))))

;;;----------------------------------------------------------------------
;;; Procedure compile-stmt-preparemission:
;;;----------------------------------------------------------------------

(define (compile-stmt-preparemission error-exit state active-symtab full-symtab stmt target)

  (let ((linenumber     (cadr stmt))
	(launcher       (caddr stmt))
	(launchitemlist (cadddr stmt))
	(stmtlist (cadr (cdddr stmt))))

    (if (state-in-display? state)
	(compile-error error-exit linenumber "prepare/mission statement cannot be used within a display"))
    
    (let ((sym (search-symtab active-symtab launcher)))

      (cond ((not sym)
	     (compile-error error-exit linenumber (string-append "unknown variable \"" (symbol->string launcher) "\"")))

	    ((or (sym-static-statesafe-launcher? sym) (sym-auto-statesafe-launcher? sym))
	     (call-with-values (lambda () (compile-stmtlist error-exit
							    state
							    active-symtab
							    full-symtab
							    stmtlist
							    target))
	       (lambda (new-state
			new-active-symtab
			new-full-symtab
			new-instructions
			new-post-state-instructions
			new-post-control-instructions
			new-subroutine-instructions)

		 (let ((compiled-launch-items (map (lambda (launchitem)
						     (let ((temp (gen-temp "discard")))
						       (call-with-values (lambda () (compile-stmtlist error-exit
												      state
												      active-symtab
												      new-full-symtab
												      (cdr launchitem)
												      `(temp ,temp)))
							 (lambda (item-state
								  item-active-symtab
								  item-full-symtab
								  item-instructions
								  item-post-state-instructions
								  item-post-control-instructions
								  item-subroutine-instructions)

							   (set! new-full-symtab item-full-symtab)
							   (set! new-subroutine-instructions `(,@new-subroutine-instructions
											       ,@item-subroutine-instructions))

							   item-instructions))))
												    
						   launchitemlist)))
						   
		 (values state
			 active-symtab
			 new-full-symtab

			 `(,@(let ((mission-id (gen-mission-id))
				   (launcher-stackframe-temp (gen-temp "launcherstackframe"))
				   (mission-stackframe-temp (gen-temp "missionstackframe"))
				   (mission-finished-label (gen-label "missionfinished")))
			       
			       `((comment ,(string-append "Prepare launcher \"" (symbol->string launcher) "\" and perform mission " (number->string mission-id) ":"))

				 (declare-temp ,launcher-stackframe-temp)
				 (declare-temp ,mission-stackframe-temp)
				 
				 ;; Determine which stack frame the launcher variable is located
				 ;; in, and clear it.

				 ,(if (sym-static-statesafe-launcher? sym)
				      `(assign (temp ,launcher-stackframe-temp) (const 0))
				      `(stack-maxindex (temp ,launcher-stackframe-temp)))

				 (delete-var (temp ,launcher-stackframe-temp) (const ,(sym-statesafe-index sym)))

				 ;; Write the current (mission) stackframe into slot 0 of the
				 ;; launcher.

				 (stack-maxindex (temp ,mission-stackframe-temp))
				 (write-array-element (temp ,mission-stackframe-temp)
						      (temp ,launcher-stackframe-temp)
						      (const ,(sym-statesafe-index sym))
						      (const 1)
						      (const 0)
						      (const 0))
						
				 ;; Clear slot 1 of the launcher (for storing the launch
				 ;; selection later).

				 (write-array-element (const 0)
						      (temp ,launcher-stackframe-temp)
						      (const ,(sym-statesafe-index sym))
						      (const 1)
						      (const 1)
						      (const 0))

				 ;; Write the mission id to slot 2 of the launcher. This will be
				 ;; used to specify the location to jump to after a launch
				 ;; selection is made.

				 (write-array-element (const ,mission-id)
						      (temp ,launcher-stackframe-temp)
						      (const ,(sym-statesafe-index sym))
						      (const 1)
						      (const 2)
						      (const 0))

				 ;; Write the launcher labels to slots 3+ of the launcher.

				 ,@(let ((instructions '()))
				     (do ((i 0 (+ i 1)))
					 ((= i (length launchitemlist)) instructions)
				       (set! instructions `(,@instructions
							    (write-array-element (const ,(car (list-ref launchitemlist i)))
										 (temp ,launcher-stackframe-temp)
										 (const ,(sym-statesafe-index sym))
										 (const 1)
										 (const ,(+ i 3))
										 (const 0))))))

				 (comment "Launcher is ready to go. Execute mission instructions.")
				 ,@new-instructions

				 (comment "Finished mission instructions. Jump to the end.")
				 (goto ,mission-finished-label)

				 (launch-label (const ,mission-id))

				 (comment "Launcher selection will trigger non-local jump to this point. Reset stack and execute selection.")
				 
				 ;; At this point we've executed the mission, which may have included displays. We have to assume
				 ;; that all temporary variables have forgotten their values. They are still declared though, so we
				 ;; don't have to generate new temps.

				 ,(if (sym-static-statesafe-launcher? sym)
				      `(assign (temp ,launcher-stackframe-temp) (const 0))
				      `(stack-maxindex (temp ,launcher-stackframe-temp)))

				 ;; There's a potential problem here - if the launcher is an automatic variable (i.e. local to a
				 ;; subroutine) and the launch selection was triggered while nested in a few subroutines, then we
				 ;; may have to reach up a few stack frames to access the launcher variable. There's no way to do
				 ;; that at the moment.
				 ;; 
				 ;; I'm going to ignore this problem, and assume that the launcher is either a global variable, or
				 ;; it's automatic but the launch selection was triggered from this same stack frame. That's safe
				 ;; because right now there is no way for another procedure to refer to this launcher unless it's a
				 ;; global variable. We can't pass it as a subroutine argument yet, which is a good thing until we
				 ;; sort out this larger problem.
				 ;; 
				 ;; Anyway, getting back to the job at hand - the next step is to truncate the stack to the size it
				 ;; was when the mission was launched.

				 ,@(let ((temp (gen-temp "missionstackframe")))
				     `((declare-temp ,temp)
				       (read-array-element (temp ,temp)
							   (temp ,launcher-stackframe-temp)
							   (const ,(sym-statesafe-index sym))
							   (const 1)
							   (const 0)
							   (const 0))
				       (stack-truncate (temp ,temp))))

				 (comment "Read launch selection.")
				 
				 ,@(let ((selection-temp (gen-temp "launchselection"))
					 (cmp-temp (gen-temp "cmp")))

				     `((declare-temp ,selection-temp)
				       (declare-temp ,cmp-temp)

				       (read-array-element (temp ,selection-temp)
							   (temp ,launcher-stackframe-temp)
							   (const ,(sym-statesafe-index sym))
							   (const 1)
							   (const 1)
							   (const 0))

				       ,@(let ((instructions '()))
					   (do ((i 0 (+ i 1)))
					       ((= i (length launchitemlist)) instructions)
					     (let ((next-label (gen-label "next")))
					       (set! instructions `(,@instructions

								    (comment ,(string-append "Check if launch selection is \""
											     (car (list-ref launchitemlist i))
											     "\", and if so execute handler."))

								    (number-equal (temp ,cmp-temp) (temp ,selection-temp) (const ,(+ i 3)))
								    (not (temp ,cmp-temp) (temp ,cmp-temp))
								    (goto-if-true ,next-label (temp ,cmp-temp))

								    ;; if we are here, then we found a matching launch item - interpolate
								    ;; the compiled code at this point

								    ,@(list-ref compiled-launch-items i)

								    (goto ,mission-finished-label)
								    ,next-label)))))
								    
				       (comment "None of the handlers matched. Display an error message. (This should never happen.)")
				       (display ,(gen-display-id) "internal error" normal
						(html () (const "<p>Internal error - launcher selection not recognized.</p>")))))

				 ,mission-finished-label
				 (delete-var (temp ,launcher-stackframe-temp) (const ,(sym-statesafe-index sym)))
				 (comment ,(string-append "Finished mission " (number->string mission-id) ".")))))
			 '()
			 '()
			 new-subroutine-instructions)))))

	    (else (compile-error error-exit linenumber (string-append "\"" (symbol->string launcher) "\" is not of type launcher")))))))

;;;----------------------------------------------------------------------
;;; Predicates on expressions:
;;;----------------------------------------------------------------------

(define (expr-self-evaluating? expr)
  (or (string? expr) (number? expr)))

(define (expr-variable? expr)
  (symbol? expr))

(define (expr-application? expr)
  (pair? expr))

;;;----------------------------------------------------------------------
;;; Procedures to compile expressions:
;;;----------------------------------------------------------------------

(define (compile-expr-general error-exit linenumber state active-symtab full-symtab expr target)

  (cond ((expr-self-evaluating? expr)
	 (compile-expr-general-self-evaluating error-exit linenumber state active-symtab full-symtab expr target))

	((expr-variable? expr)
	 (let ((sym (search-symtab active-symtab expr)))

	   (cond ((not sym)
		  (compile-error error-exit linenumber (string-append "unknown variable \"" (symbol->string expr) "\"")))

		 ((sym-static-statesafe-var? sym)
		  (compile-expr-general-static-statesafe-var error-exit linenumber state active-symtab full-symtab sym target))

		 ((sym-auto-statesafe-var? sym)
		  (compile-expr-general-auto-statesafe-var error-exit linenumber state active-symtab full-symtab sym target))

		 ((or (sym-static-statesafe-table? sym) (sym-auto-statesafe-table? sym))
		  (compile-error error-exit linenumber "encountered table variable in expression"))

		 ;; For now, compile enum like a var for evaluation purposes.
		 ((sym-static-statesafe-enum? sym)
		  (compile-expr-general-static-statesafe-var error-exit linenumber state active-symtab full-symtab sym target))

		 ;; For now, compile enum like a var for evaluation purposes.
		 ((sym-auto-statesafe-enum? sym)
		  (compile-expr-general-auto-statesafe-var error-exit linenumber state active-symtab full-symtab sym target))

		 ((or (sym-static-transient-var? sym) (sym-auto-transient-var? sym))
		  (compile-expr-general-transient-var error-exit linenumber state active-symtab full-symtab sym target))

		 (else
		  (compile-error error-exit linenumber (string-append "variable \"" (symbol->string expr) "\" has invalid or unimplemented type"))))))

	((expr-application? expr)
	 (compile-expr-general-application error-exit linenumber state active-symtab full-symtab expr target))

	(else
	 (compile-error error-exit "invalid expression " expr))))

(define (compile-expr-general-self-evaluating error-exit linenumber state active-symtab full-symtab expr target)
  `((assign ,target (const ,expr))))

(define (compile-expr-general-static-statesafe-var error-exit linenumber state active-symtab full-symtab sym target)
  `((read-scalar ,target (const 0) (const ,(sym-statesafe-index sym)))))

(define (compile-expr-general-auto-statesafe-var error-exit linenumber state active-symtab full-symtab sym target)
  (let ((stackframe-temp (gen-temp "stackframe")))
    `((declare-temp ,stackframe-temp)
      (stack-maxindex (temp ,stackframe-temp))
      (read-scalar ,target (temp ,stackframe-temp) (const ,(sym-statesafe-index sym))))))

(define (compile-expr-general-transient-var error-exit linenumber state active-symtab full-symtab sym target)
  `((assign ,target (temp ,(sym-transient-varname sym)))))

(define (compile-expr-general-application error-exit linenumber state active-symtab full-symtab expr target)
  (let ((function (car expr)))

    (cond ((eq? function '+)
	   (let ((temp (gen-temp "operand")))
	     `((declare-temp ,temp)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp))  ; put RHS in temp
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) target)          ; assign LHS to target
	       (increment ,target (temp ,temp)))))                                 ; increment target by RHS                             

	  ((eq? function '-)
	   (let ((temp (gen-temp "operand")))
	     `((declare-temp ,temp)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp))
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) target)
	       (decrement ,target (temp ,temp)))))

	  ((eq? function '*)
	   (let ((temp (gen-temp "operand")))
	     `((declare-temp ,temp)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp))
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) target)
	       (multiply ,target (temp ,temp)))))

	  ((eq? function '/)
	   (let ((temp (gen-temp "operand")))
	     `((declare-temp ,temp)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp))
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) target)
	       (divide ,target (temp ,temp)))))

	  ((eq? function 'sqr)
	   (let ((temp1 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (sqrt ,target (temp ,temp1)))))

	  ((eq? function '==)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (number-equal ,target (temp ,temp1) (temp ,temp2)))))

	  ((eq? function '!=)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (number-equal ,target (temp ,temp1) (temp ,temp2))
	       (not ,target ,target))))

	  ((eq? function '<)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (number-less ,target (temp ,temp1) (temp ,temp2)))))

	  ((eq? function '>)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (number-greater ,target (temp ,temp1) (temp ,temp2)))))

	  ((eq? function '<=)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (number-greater ,target (temp ,temp1) (temp ,temp2))
	       (not ,target ,target))))

	  ((eq? function '>=)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (number-less ,target (temp ,temp1) (temp ,temp2))
	       (not ,target ,target))))

	  ((eq? function 'eq)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (string-equal ,target (temp ,temp1) (temp ,temp2)))))

	  ((eq? function 'ne)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (string-equal ,target (temp ,temp1) (temp ,temp2))
	       (not ,target ,target))))

	  ((eq? function 'and)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (and ,target (temp ,temp1) (temp ,temp2)))))

	  ((eq? function 'or)
	   (let ((temp1 (gen-temp "operand"))
		 (temp2 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (declare-temp ,temp2)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (caddr expr) `(temp ,temp2))
	       (or ,target (temp ,temp1) (temp ,temp2)))))

	  ((eq? function 'not)
	   (let ((temp1 (gen-temp "operand")))
	     `((declare-temp ,temp1)
	       ,@(compile-expr-general error-exit linenumber state active-symtab full-symtab (cadr expr) `(temp ,temp1))
	       (not ,target (temp ,temp1)))))

	  ((eq? function 'external)
	   (compile-expr-general-external-subroutine-call error-exit linenumber state active-symtab full-symtab expr target))

	  (else
	   (compile-expr-general-application-subroutine-call error-exit linenumber state active-symtab full-symtab expr target)))))

(define (compile-expr-general-external-subroutine-call error-exit linenumber state active-symtab full-symtab expr target)

  (if (state-in-display? state)
      (compile-error error-exit linenumber "cannot call external subroutine within display"))

  (let ((subroutine-name (caddr expr))
	(subroutine-args (cdddr expr)))

    (let ((instructions '((comment "Evaluating arguments in preparation for (external) subroutine call:")))
	  (evaluated-args '()))

      (for-each (lambda (arg)
		  (let ((temp (gen-temp "arg")))
		    (let ((new-instructions (compile-expr-general error-exit linenumber state active-symtab full-symtab arg `(temp ,temp))))
		      (set! instructions `(,@instructions
					   (declare-temp ,temp)
					   ,@new-instructions))
		      (set! evaluated-args `(,@evaluated-args (temp ,temp))))))
		subroutine-args)
		     
      `(,@instructions
	(external ,target ,subroutine-name ,@evaluated-args)))))

(define (compile-expr-general-application-subroutine-call error-exit linenumber state active-symtab full-symtab expr target)

  (if (state-in-display? state)
      (compile-error error-exit linenumber "cannot call subroutine within display"))

  (let ((subroutine-name (car expr))
	(subroutine-args (cdr expr)))

    (let ((sym (search-symtab active-symtab subroutine-name)))
      
      (cond ((not sym)
	     (compile-error error-exit linenumber (string-append "unknown subroutine \"" (symbol->string subroutine-name) "\"")))

	    ((sym-subroutine? sym)
	     (let ((subroutine-id (sym-subroutine-id sym))
		   (subroutine-label (sym-subroutine-label sym)))

	       (let ((instructions '((comment "Evaluating arguments in preparation for subroutine call:")))
		     (evaluated-args '()))

		 (for-each (lambda (arg)
			     (let ((temp (gen-temp "arg")))
			       (let ((new-instructions (compile-expr-general error-exit linenumber state active-symtab full-symtab arg `(temp ,temp))))
				 (set! instructions `(,@instructions
						      (declare-temp ,temp)
						      ,@new-instructions))
				 (set! evaluated-args `(,@evaluated-args (temp ,temp))))))
			   subroutine-args)
		 
		 `(,@instructions
		   (gosub ,target ,subroutine-name ,subroutine-id ,subroutine-label ,(gen-call) ,@evaluated-args)))))

	    (else
	     (compile-error error-exit linenumber (string-append "\"" (symbol->string subroutine-name) "\" is not a subroutine")))))))

;;;----------------------------------------------------------------------
;;; Procedures that operate on the compiler state:
;;;----------------------------------------------------------------------

(define (new-state)
  '())

(define (extend-state state item)
  (cons item state))

(define (state-in-display display-id)
  `(in-display ,display-id))

(define (state-in-subroutine subroutine-id name)
  `(in-subroutine ,subroutine-id ,name))

(define (state-in-loop endlabel)
  `(in-loop ,endlabel))

(define (state-in-conditional)
  `(in-conditional))

(define (state-in-display? state)
  (assq 'in-display state))

(define (state-in-subroutine? state)
  (assq 'in-subroutine state))

(define (state-in-loop? state)
  (assq 'in-loop state))

(define (state-in-conditional? state)
  (assq 'in-conditional state))

(define (state-display-id state)
  (let ((item (assq 'in-display state)))
    (if item
	(cadr item)
	#f)))

(define (state-subroutine-id state)
  (let ((item (assq 'in-subroutine state)))
    (if item
	(cadr item)
	#f)))

(define (state-subroutine-name state)
  (let ((item (assq 'in-subroutine state)))
    (if item
	(caddr item)
	#f)))

(define (state-loop-endlabel state)
  (let ((item (assq 'in-loop state)))
    (if item
	(cadr item)
	#f)))

;;;----------------------------------------------------------------------
;;; Procedures that operate on the symbol table:
;;;----------------------------------------------------------------------

(define (new-symtab)
  '())

(define (extend-symtab symtab sym)
  (cons sym symtab))

(define (search-symtab symtab sym)
  (assq sym symtab))

;;;----------------------------------------------------------------------
;;; Procedures that operate on symbol table entries for statesafe variables:
;;;----------------------------------------------------------------------

(define (new-sym-static-statesafe name type)
  `(,name static-statesafe ,(gen-static-statesafe) 0 "<main>" ,@type))

(define (new-sym-auto-statesafe subroutine-id subroutine-name name type)
  `(,name auto-statesafe ,(gen-auto-statesafe subroutine-id) ,subroutine-id ,subroutine-name ,@type))

(define (sym-static-statesafe-var? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'static-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (eq? (cadddr (cdddr sym)) 'var)))
  
(define (sym-static-statesafe-table? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'static-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (pair? (cadddr (cdddr sym)))
       (eq? (car (cadddr (cdddr sym))) 'table)))
  
(define (sym-static-statesafe-enum? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'static-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (pair? (cadddr (cdddr sym)))
       (eq? (car (cadddr (cdddr sym))) 'enum)))
  
(define (sym-static-statesafe-launcher? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'static-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (eq? (cadddr (cdddr sym)) 'launcher)))

(define (sym-static-statesafe-text? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'static-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (eq? (cadddr (cdddr sym)) 'text)))

(define (sym-auto-statesafe-var? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'auto-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (eq? (cadddr (cdddr sym)) 'var)))
  
(define (sym-auto-statesafe-table? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'auto-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (pair? (cadddr (cdddr sym)))
       (eq? (car (cadddr (cdddr sym))) 'table)))
  
(define (sym-auto-statesafe-enum? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'auto-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (pair? (cadddr (cdddr sym)))
       (eq? (car (cadddr (cdddr sym))) 'enum)))

(define (sym-auto-statesafe-launcher? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'auto-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (eq? (cadddr (cdddr sym)) 'launcher)))

(define (sym-auto-statesafe-text? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'auto-statesafe)
       (eq? (caddr (cdddr sym)) 'statesafe)
       (eq? (cadddr (cdddr sym)) 'text)))

(define (sym-statesafe-index sym)
  (caddr sym))

(define (sym-statesafe-subroutine-id sym)
  (cadddr sym))

(define (sym-statesafe-subroutine-name sym)
  (cadr (cdddr sym)))

(define (sym-statesafe-type-fields sym)
  (cdr (cadddr (cdddr sym))))

;;;----------------------------------------------------------------------
;;; Procedures that operate on symbol table entries for transient variables:
;;;----------------------------------------------------------------------

(define (new-sym-static-transient name type)
  `(,name static-transient ,(gen-temp (symbol->string name)) 0 "<main>" ,@type))

(define (new-sym-auto-transient subroutine-id subroutine-name name type)
  `(,name auto-transient ,(gen-temp (symbol->string name)) ,subroutine-id ,subroutine-name ,@type))

(define (sym-static-transient-var? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'static-transient)
       (eq? (caddr (cdddr sym)) 'transient)
       (eq? (cadddr (cdddr sym)) 'var)))

(define (sym-auto-transient-var? sym)
  (and (= (length sym) 7)
       (eq? (cadr sym) 'auto-transient)
       (eq? (caddr (cdddr sym)) 'transient)
       (eq? (cadddr (cdddr sym)) 'var)))

(define (sym-transient-varname sym)
  (caddr sym))

(define (sym-transient-subroutine-id sym)
  (cadddr sym))

(define (sym-transient-subroutine-name sym)
  (cadr (cdddr sym)))

;;;----------------------------------------------------------------------
;;; Procedures that operate on symbol table entries for subroutine names:
;;;----------------------------------------------------------------------

(define (new-sym-subroutine name)
  `(,name subroutine ,(gen-subroutine-id) ,(gen-label (symbol->string name))))

(define (sym-subroutine? sym)
  (and (= (length sym) 4)
       (eq? (cadr sym) 'subroutine)))

(define (sym-subroutine-id sym)
  (caddr sym))

(define (sym-subroutine-label sym)
  (cadddr sym))

;;;----------------------------------------------------------------------
;;; Procedures that operate on symbol table entries for typedefs:
;;;----------------------------------------------------------------------

(define (new-sym-typedef name type)
  `(,name typedef ,type))

(define (sym-typedef? sym)
  (and (= (length sym) 3)
       (eq? (cadr sym) 'typedef)))

(define (sym-typedef-type sym)
  (caddr sym))

(define (search-symtab-type symtab type error-exit linenumber)
  (cond ((symbol? type)
	 (if (or (eq? type 'var) (eq? type 'launcher) (eq? type 'text))
	     type
	     (let ((sym (search-symtab symtab type)))

	       (cond ((not sym)
		      (compile-error error-exit linenumber "unknown type"))

		     ((sym-typedef? sym)
		      (sym-typedef-type sym))

		     (else
		      (compile-error error-exit linenumber (string-append "\"" (symbol->string type) "\" is not a type")))))))

	((pair? type)
	 (if (or (eq? (car type) 'table) (eq? (car type) 'enum))
	     type
	     (compile-error error-exit linenumber "bad type")))

	(else
	 (compile-error error-exit linenumber "bad type"))))

;;;----------------------------------------------------------------------
;;; Procedures that operate on all types of symbol table entries:
;;;----------------------------------------------------------------------

(define (sym-name sym)
  (car sym))

;;;----------------------------------------------------------------------
;;; Procedures that operate on type fields:
;;;----------------------------------------------------------------------

(define (fieldlist-colcount fields)
  (+ (length fields) 1))

(define (field-type field)
  (car field))

(define (field-name field)
  (cadr field))

(define (field-label field)
  (if (pair? (cddr field))
      (caddr field)
      (symbol->string (field-name field))))

;;;----------------------------------------------------------------------
;;; Procedures to extract symbol table information needed by back end:
;;;----------------------------------------------------------------------

(define (symtab-schema-entries symtab)
  (let ((entries '()))
    (for-each (lambda (sym)
		(cond ((or (sym-static-statesafe-var? sym) (sym-auto-statesafe-var? sym))
		       (set! entries `(,@entries
				       ((variable ,(sym-statesafe-index sym))
					(subroutine-id ,(sym-statesafe-subroutine-id sym))
					(variable-name ,(sym-name sym))
					(type-name "var")
					(type-index 0)
					(field-name #f)
					(field-type #f)
					(field-label #f)))))

		      ((or (sym-static-statesafe-text? sym) (sym-auto-statesafe-text? sym))
		       (set! entries `(,@entries
				       ((variable ,(sym-statesafe-index sym))
					(subroutine-id ,(sym-statesafe-subroutine-id sym))
					(variable-name ,(sym-name sym))
					(type-name "text")
					(type-index 0)
					(field-name #f)
					(field-type #f)
					(field-label #f)))))

		      ((or (sym-static-statesafe-table? sym) (sym-auto-statesafe-table? sym))
		       (let ((fields (sym-statesafe-type-fields sym)))
			 (do ((i 0 (+ i 1)))
			     ((= i (length fields)) #f)
			   (let ((field (list-ref fields i)))
			     (set! entries `(,@entries
					     ((variable ,(sym-statesafe-index sym))
					      (subroutine-id ,(sym-statesafe-subroutine-id sym))
					      (variable-name ,(sym-name sym))
					      (type-name "table")
					      (type-index ,i)
					      (field-name ,(field-name field))
					      (field-type ,(field-type field))
					      (field-label ,(field-label field)))))))))

		      ((or (sym-static-statesafe-enum? sym) (sym-auto-statesafe-enum? sym))
		       (let ((fields (sym-statesafe-type-fields sym)))
			 (do ((i 0 (+ i 1)))
			     ((= i (length fields)) #f)
			   (let ((field (list-ref fields i)))
			     (set! entries `(,@entries
					     ((variable ,(sym-statesafe-index sym))
					      (subroutine-id ,(sym-statesafe-subroutine-id sym))
					      (variable-name ,(sym-name sym))
					      (type-name "enum")
					      (type-index ,i)
					      (field-name ,(field-name field))
					      (field-type ,(field-type field))
					      (field-label ,(field-label field)))))))))

		      ((or (sym-static-statesafe-launcher? sym) (sym-auto-statesafe-launcher? sym))
		       (set! entries `(,@entries
				       ((variable ,(sym-statesafe-index sym))
					(subroutine-id ,(sym-statesafe-subroutine-id sym))
					(variable-name ,(sym-name sym))
					(type-name "launcher")
					(type-index 0)
					(field-name #f)
					(field-type #f)
					(field-label #f)))))

		      ((sym-subroutine? sym)
		       (set! entries `(,@entries
				       ((variable ,(sym-subroutine-id sym))
					(subroutine-id 2)                      ; subroutine_id is always 2 for subroutine name->id mappings
					(variable-name ,(sym-name sym))
					(type-name "subroutine")
					(type-index 0)
					(field-name #f)
					(field-type #f)
					(field-label #f)))))))
	      symtab)
    entries))

(define (display-schema-entries instructions)
  (let ((entries '()))
    (for-each (lambda (i)
		(if (and (pair? i) (eq? (car i) 'display))
		    (let ((display-id (cadr i))
			  (display-name (caddr i)))
		      (set! entries `(,@entries
				      ((variable ,display-id)
				       (subroutine-id 1)         ; subroutine_id is always 1 for displays
				       (variable-name ,display-name)
				       (type-name "display")
				       (type-index 0)
				       (field-name #f)
				       (field-type #f)
				       (field-label #f)))))))
	      instructions)
    entries))

(define (symtab-transient-vars symtab)
  (let ((vars '()))
    (for-each (lambda (sym)
		(if (or (sym-static-transient-var? sym) (sym-auto-transient-var? sym))
		    (set! vars `(,@vars ,(sym-transient-varname sym)))))
	      symtab)
    vars))

;;;----------------------------------------------------------------------
;;; Procedures to generate labels, temporary variables, and so on:
;;;----------------------------------------------------------------------

(define gen-static-statesafe
  (let ((statesafe-variable-current-index 0))
    (lambda ()
      (let ((statesafe-index statesafe-variable-current-index))
	(set! statesafe-variable-current-index (+ statesafe-variable-current-index 1))
	statesafe-index))))

(define gen-auto-statesafe
  (let ((current-subroutine-id 0)
	(statesafe-variable-current-index 0))
    (lambda (subroutine-id)
      (if (not (= subroutine-id current-subroutine-id))
	  (begin
	    (set! statesafe-variable-current-index 3)
	    (set! current-subroutine-id subroutine-id)))
      (let ((statesafe-index statesafe-variable-current-index))
	(set! statesafe-variable-current-index (+ statesafe-variable-current-index 1))
	statesafe-index))))

(define gen-temp
  (let ((transient-variable-current-index 0))
    (lambda (name-template)
      (let ((tempname (string-append "t" (number->string transient-variable-current-index) "_" name-template)))
	(set! transient-variable-current-index (+ transient-variable-current-index 1))
	tempname))))

(define gen-label
  (let ((label-current-index 0))
    (lambda (label-template)
      (let ((labelname (string-append "l" (number->string label-current-index) "_" label-template)))
	(set! label-current-index (+ label-current-index 1))
	labelname))))

(define gen-action-selector-value
  (let ((param-current-index 0))
    (lambda (param-template)
      (let ((param (string-append param-template "-" (number->string param-current-index) "-")))
	(set! param-current-index (+ param-current-index 1))
	param))))

(define gen-subroutine-id
  (let ((subroutine-current-index 64))  ; subroutine numbers lower than 64 are reserved
    (lambda ()
      (let ((sub-id subroutine-current-index))
	(set! subroutine-current-index (+ subroutine-current-index 1))
	sub-id))))

(define gen-call
  (let ((subroutine-call-current-index 0))
    (lambda ()
      (let ((call-id subroutine-call-current-index))
	(set! subroutine-call-current-index (+ subroutine-call-current-index 1))
	call-id))))

(define gen-mission-id
  (let ((mission-current-index 0))
    (lambda ()
      (let ((mission-id mission-current-index))
	(set! mission-current-index (+ mission-current-index 1))
	mission-id))))

(define gen-display-id
  (let ((display-current-index 3))  ; displays 0, 1, and 2 are reserved
    (lambda ()
      (let ((display-id display-current-index))
	(set! display-current-index (+ display-current-index 1))
	display-id))))

(define (gen-schema-id)
  (string-append (number->string (getpid))
		 "-"
		 (number->string (current-time))
		 "-"
		 (number->string (random 1000000))
		 "-"
		 (number->string (random 1000000))
		 "-"
		 (number->string (random 1000000))))

;;;----------------------------------------------------------------------
;;; End of procedure definitions. Run the main program.
;;;----------------------------------------------------------------------

(main)
